from .animate import create_animation
from .common import (
    DEFAULT_QUANTITIES,
    PLOT_DEFAULTS,
    Quantity,
    auto_grid,
    auto_range_iqr,
)
from .interactive import InteractivePlot, MultiInteractivePlot



from typing import Sequence, Tuple

import h5py
import numpy as np

from ..finite_volume.mesh import Mesh
from ..finite_volume.operators import build_gradient
from ..solution.data import TDGLData, get_edge_quantity_data, load_state_data
from .common import Quantity


def get_plot_data(
    h5file: h5py.File, mesh: Mesh, quantity: Quantity, frame: int
) -> Tuple[np.ndarray, np.ndarray, Sequence[float]]:
    """Get data to plot.

    Args:
        h5file: The data file.
        mesh: The mesh used in the simulation.
        quantity: The quantity to return.
        frame: The current frame.

    Returns:
        A tuple of the values for the color plot, the directions for the
        quiver plot and the limits for the color plot.
    """
    tdgl_data = TDGLData.from_hdf5(h5file, frame)
    psi = tdgl_data.psi
    mu = tdgl_data.mu
    a_applied = tdgl_data.applied_vector_potential
    a_induced = tdgl_data.induced_vector_potential
    supercurrent = tdgl_data.supercurrent
    normal_current = tdgl_data.normal_current
    nsites = len(mesh.sites)

    if quantity is Quantity.ORDER_PARAMETER and psi is not None:
        return np.abs(psi), np.zeros((nsites, 2)), [0, 1]

    elif quantity is Quantity.PHASE and psi is not None:
        return np.angle(psi) / np.pi, np.zeros((nsites, 2)), [-1, 1]

    elif quantity is Quantity.SUPERCURRENT and supercurrent is not None:
        return get_edge_quantity_data(supercurrent, mesh)

    elif quantity is Quantity.NORMAL_CURRENT and normal_current is not None:
        return get_edge_quantity_data(normal_current, mesh)

    elif quantity is Quantity.SCALAR_POTENTIAL and mu is not None:
        mu = mu - np.nanmin(mu)
        return mu, np.zeros((nsites, 2)), [np.min(mu), np.max(mu)]

    elif quantity is Quantity.APPLIED_VECTOR_POTENTIAL and a_applied is not None:
        return get_edge_quantity_data(
            (a_applied * mesh.edge_mesh.directions).sum(axis=1), mesh
        )

    elif quantity is Quantity.INDUCED_VECTOR_POTENTIAL and a_induced is not None:
        return get_edge_quantity_data(
            (a_induced * mesh.edge_mesh.directions).sum(axis=1), mesh
        )

    elif quantity is Quantity.EPSILON:
        if "epsilon" in h5file:
            epsilon = np.asarray(h5file["epsilon"])
        else:
            epsilon = np.ones_like(mu)
        return epsilon, np.zeros((nsites, 2)), [np.min(epsilon), np.max(epsilon)]

    elif (
        quantity is Quantity.VORTICITY
        and supercurrent is not None
        and normal_current is not None
    ):
        j_sc_site = mesh.get_quantity_on_site(supercurrent)
        j_nm_site = mesh.get_quantity_on_site(normal_current)
        j_site = j_sc_site + j_nm_site
        gradient = build_gradient(mesh)
        normalized_directions = (
            mesh.edge_mesh.directions
            / np.linalg.norm(mesh.edge_mesh.directions, axis=1)[:, np.newaxis]
        )
        grad_jx = gradient @ j_site[:, 0]
        grad_jy = gradient @ j_site[:, 1]
        djy_dx = grad_jy * normalized_directions[:, 0]
        djx_dy = grad_jx * normalized_directions[:, 1]
        vorticity_on_edges = djy_dx - djx_dy
        vorticity = mesh.get_quantity_on_site(vorticity_on_edges, vector=False)
        vmax = max(np.abs(np.max(vorticity)), np.abs(np.min(vorticity)))
        return vorticity, np.zeros((nsites, 2)), [-vmax, vmax]

    return np.zeros(nsites), np.zeros((nsites, 2)), [0, 0]


def get_state_string(h5file: h5py.File, frame: int, max_frame: int) -> str:
    state = load_state_data(h5file, frame)

    state_string = f"Frame {frame} of {max_frame}"
    i = 1
    for key, value in state.items():
        if key == "timestamp":
            continue
        state_string += ", "
        if i % 3 == 0:
            state_string += "\n"
        if type(value) is np.float64:
            state_string += f"{key}: {value:.2e}"
        else:
            state_string += f"{key}: {value}"

        i += 1

    return state_string



import os
import tempfile

import h5py
import numpy as np
import pint
import pytest

import tdgl
from tdgl.solution.data import DynamicsData


@pytest.fixture(scope="module")
def tempdir():
    tmp = tempfile.TemporaryDirectory()
    yield tmp.__enter__()
    tmp.cleanup()


@pytest.fixture(scope="module")
def solution(transport_device, tempdir):
    device = transport_device
    total_time = 100

    fname = os.path.join(tempdir, "output.h5")
    options = tdgl.SolverOptions(
        solve_time=total_time,
        output_file=fname,
        save_every=100,
        field_units="uT",
        current_units="uA",
    )
    field = tdgl.sources.ConstantField(1)

    def terminal_currents(time):
        return dict(source=0.1, drain=-0.1)

    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=field,
        terminal_currents=terminal_currents,
    )

    _, phases = solution.boundary_phases()[device.film.name]
    assert np.isclose((phases[-1] - phases[0]) / (2 * np.pi), 0, atol=5e-2)

    _, phases = solution.boundary_phases(delta=True)[device.film.name]
    assert np.isclose(phases[-1] / (2 * np.pi), 0, atol=5e-2)
    return solution


def test_save_and_load_solution(solution, tempdir):
    path = os.path.join(tempdir, "output-1.h5")
    solution.to_hdf5(path)
    loaded_solution = tdgl.Solution.from_hdf5(path)
    assert loaded_solution == solution

    solution.to_hdf5()
    loaded_solution = tdgl.Solution.from_hdf5(solution.path)
    assert loaded_solution == solution


@pytest.mark.parametrize("hole", ["hole1", "hole2", "invalid"])
@pytest.mark.parametrize("with_units", [False, True])
@pytest.mark.parametrize("units", ["Phi_0", "mT * um**2"])
@pytest.mark.parametrize("interp_method", ["linear", "cubic", "nearest"])
def test_hole_fluxoid(solution, hole, with_units, units, interp_method):
    if hole == "invalid":
        with pytest.raises(KeyError):
            _ = solution.hole_fluxoid(
                hole, units=units, with_units=with_units, interp_method=interp_method
            )
    else:
        fluxoid = solution.hole_fluxoid(
            hole, units=units, with_units=with_units, interp_method=interp_method
        )
        assert isinstance(fluxoid, tdgl.Fluxoid)
        if with_units:
            assert isinstance(sum(fluxoid), pint.Quantity)
        else:
            assert isinstance(sum(fluxoid), float)


def test_tdgl_data(solution: tdgl.Solution):
    tdgl_data = solution.tdgl_data
    with h5py.File(solution.path, "r+") as f:
        tdgl_data.to_hdf5(f.create_group("tdgl_data"))


def test_dynamics(solution: tdgl.Solution):
    dynamics = solution.dynamics
    ix = dynamics.time_slice()
    ts = dynamics.time[ix]
    assert np.array_equal(ts, dynamics.time)

    ix = dynamics.time_slice(tmin=10, tmax=90)
    ts = dynamics.time[ix]
    assert np.all(ts >= 10)
    assert np.all(ts <= 90)

    V0 = dynamics.mean_voltage()
    d2 = dynamics.resample()
    dt = np.diff(d2.time)
    assert np.allclose(dt, dt[0])
    V1 = d2.mean_voltage()

    assert np.isclose(V0, V1, rtol=1e-2)

    with tdgl.non_gui_backend():
        _ = dynamics.plot(legend=True)
        _ = dynamics.plot_dt()

    with h5py.File(solution.path, "r+") as f:
        dynamics.to_hdf5(f.create_group("dynamics"))
        loaded_dynamics = DynamicsData.from_hdf5(f["dynamics"])

    assert loaded_dynamics == dynamics

    time = solution.times
    assert len(time) == (solution.data_range[1] + 1)
    assert solution.closest_solve_step(0) == 0



import inspect
import operator
from typing import Callable, Optional, Union

import cloudpickle
import numpy as np


class _FakeArgSpec:
    def __init__(
        self,
        args=None,
        varargs=None,
        varkw=None,
        defaults=None,
        kwonlyargs=None,
        kwonlydefaults=None,
        annotations=None,
    ):
        self.args = args
        self.varargs = varargs
        self.varkw = varkw
        self.defaults = defaults
        self.kwonlyargs = kwonlyargs
        self.kwonlydefaults = kwonlydefaults
        self.annotations = annotations


def function_repr(
    func: Callable,
    argspec: Optional[Union[_FakeArgSpec, inspect.FullArgSpec]] = None,
) -> str:
    """Returns a human-readable string representation for a function."""
    if argspec is None:
        argspec = inspect.getfullargspec(func)
    args = [str(arg) for arg in argspec.args]

    if argspec.defaults:
        for i, val in enumerate(argspec.defaults[::-1]):
            args[-(i + 1)] = args[-(i + 1)] + f"={val!r}"

    if argspec.varargs:
        args.append("*" + argspec.varargs)

    if argspec.kwonlyargs:
        if not argspec.varargs:
            args.append("*")
        args.extend(argspec.kwonlyargs)
    if argspec.kwonlydefaults:
        for i, name in enumerate(args):
            if name in argspec.kwonlydefaults:
                args[i] = args[i] + f"={argspec.kwonlydefaults[name]!r}"
    if argspec.varkw:
        args.append("**" + argspec.varkw)

    if argspec.annotations:
        for i, name in enumerate(args):
            if name in argspec.annotations:
                args[i] = args[i] + f": {argspec.annotations[name].__name__!r}"

    return func.__name__ + "(" + ", ".join(args) + ")"


class Parameter:
    """A callable object that computes a scalar or vector quantity
    as a function of position coordinates x, y (and optionally z).

    Addition, subtraction, multiplication, and division
    between multiple Parameters and/or real numbers (ints and floats)
    is supported. The result of any of these operations is a
    ``CompositeParameter`` object.

    Args:
        func: A callable/function that actually calculates the parameter's value.
            The function must take x, y (and optionally z) as the first and only
            positional arguments, and all other arguments must be keyword arguments.
            Therefore func should have a signature like
            ``func(x, y, z, a=1, b=2, c=True)``, ``func(x, y, *, a, b, c)``,
            ``func(x, y, z, *, a, b, c)``, or ``func(x, y, z, *, a, b=None, c=3)``.
        kwargs: Keyword arguments for func.
    """

    __slots__ = ("func", "kwargs")

    def __init__(self, func: Callable, **kwargs):
        argspec = inspect.getfullargspec(func)
        args = argspec.args
        num_args = 2
        if args[:num_args] != ["x", "y"]:
            raise ValueError(
                "The first function arguments must be x and y, "
                f"not {', '.join(args[:num_args])!r}."
            )
        if "z" in args:
            if args.index("z") != num_args:
                raise ValueError(
                    "If the function takes an argument z, "
                    "it must be the third argument (x, y, z)."
                )
            num_args = 3
        defaults = argspec.defaults or []
        if len(defaults) != len(args) - num_args:
            raise ValueError(
                "All arguments other than x, y, z must be keyword arguments."
            )
        defaults_dict = dict(zip(args[num_args:], defaults))
        kwonlyargs = set(kwargs) - set(argspec.args[num_args:])
        if not kwonlyargs.issubset(set(argspec.kwonlyargs or [])):
            raise ValueError(
                f"Provided keyword-only arguments ({kwonlyargs!r}) "
                f"do not match the function signature: {function_repr(func)}."
            )
        defaults_dict.update(argspec.kwonlydefaults or {})

        self.func = func
        self.kwargs = defaults_dict
        self.kwargs.update(kwargs)

    def __call__(
        self,
        x: Union[int, float, np.ndarray],
        y: Union[int, float, np.ndarray],
        z: Optional[Union[int, float, np.ndarray]] = None,
    ) -> Union[int, float, np.ndarray]:
        kwargs = self.kwargs.copy()
        x, y = np.atleast_1d(x, y)
        if z is not None:
            kwargs["z"] = np.atleast_1d(z)
        result = self.func(x, y, **kwargs).squeeze()
        if result.ndim == 0:
            result = result.item()
        return result

    def _get_argspec(self) -> _FakeArgSpec:
        if self.kwargs:
            kwargs, kwarg_values = list(zip(*self.kwargs.items()))
        else:
            kwargs = []
            kwarg_values = []
        return _FakeArgSpec(
            args=list(kwargs),
            defaults=kwarg_values,
        )

    def __repr__(self) -> str:
        func_repr = function_repr(self.func, argspec=self._get_argspec())
        return f"{self.__class__.__name__}<{func_repr}>"

    def __add__(self, other) -> "CompositeParameter":
        """self + other"""
        return CompositeParameter(self, other, operator.add)

    def __radd__(self, other) -> "CompositeParameter":
        """other + self"""
        return CompositeParameter(other, self, operator.add)

    def __sub__(self, other) -> "CompositeParameter":
        """self - other"""
        return CompositeParameter(self, other, operator.sub)

    def __rsub__(self, other) -> "CompositeParameter":
        """other - self"""
        return CompositeParameter(other, self, operator.sub)

    def __mul__(self, other) -> "CompositeParameter":
        """self * other"""
        return CompositeParameter(self, other, operator.mul)

    def __rmul__(self, other) -> "CompositeParameter":
        """other * self"""
        return CompositeParameter(other, self, operator.mul)

    def __truediv__(self, other) -> "CompositeParameter":
        """self / other"""
        return CompositeParameter(self, other, operator.truediv)

    def __rtruediv__(self, other) -> "CompositeParameter":
        """other / self"""
        return CompositeParameter(other, self, operator.truediv)

    def __pow__(self, other) -> "CompositeParameter":
        """self ** other"""
        return CompositeParameter(self, other, operator.pow)

    def __rpow__(self, other) -> "CompositeParameter":
        """other ** self"""
        return CompositeParameter(other, self, operator.pow)

    def __eq__(self, other) -> bool:
        if other is self:
            return True

        if not isinstance(other, Parameter):
            return False

        # Check if function bytecode is the same
        if self.func.__code__ != other.func.__code__:
            return False

        return self.kwargs == other.kwargs


class CompositeParameter(Parameter):
    """A callable object that behaves like a Parameter
    (i.e. it computes a scalar or vector quantity as a function of
    position coordinates x, y, z). A CompositeParameter object is created as
    a result of mathematical operations between Parameters, CompositeParameters,
    and/or real numbers.

    Addition, subtraction, multiplication, division, and exponentiation
    between Parameters, CompositeParameters and real numbers (ints and floats)
    are supported. The result of any of these operations is a new
    CompositeParameter object.

    Args:
        left: The object on the left-hand side of the operator.
        right: The object on the right-hand side of the operator.
        operator_: The operator acting on left and right (or its string representation).
    """

    VALID_OPERATORS = {
        operator.add: "+",
        operator.sub: "-",
        operator.mul: "*",
        operator.truediv: "/",
        operator.pow: "**",
    }

    def __init__(
        self,
        left: Union[int, float, Parameter, "CompositeParameter"],
        right: Union[int, float, Parameter, "CompositeParameter"],
        operator_: Union[Callable, str],
    ):
        valid_types = (int, float, complex, Parameter, CompositeParameter)
        if not isinstance(left, valid_types):
            raise TypeError(
                f"Left must be a number, Parameter, or CompositeParameter, "
                f"not {type(left)!r}."
            )
        if not isinstance(right, valid_types):
            raise TypeError(
                f"Right must be a number, Parameter, or CompositeParameter, "
                f"not {type(right)!r}."
            )
        if isinstance(left, (int, float)) and isinstance(right, (int, float)):
            raise TypeError(
                "Either left or right must be a Parameter or CompositeParameter."
            )
        if isinstance(operator_, str):
            operators = {v: k for k, v in self.VALID_OPERATORS.items()}
            operator_ = operators.get(operator_.strip(), None)
        if operator_ not in self.VALID_OPERATORS:
            raise ValueError(
                f"Unknown operator, {operator_!r}. "
                f"Valid operators are {list(self.VALID_OPERATORS)!r}."
            )
        self.left = left
        self.right = right
        self.operator = operator_

    def __call__(
        self,
        x: Union[int, float, np.ndarray],
        y: Union[int, float, np.ndarray],
        z: Optional[Union[int, float, np.ndarray]] = None,
    ) -> Union[int, float, np.ndarray]:
        if isinstance(self.left, (int, float)):
            left_val = self.left
        else:
            left_val = self.left(x, y, z)
        if isinstance(self.right, (int, float)):
            right_val = self.right
        else:
            right_val = self.right(x, y, z)
        return self.operator(left_val, right_val)

    def _bare_repr(self) -> str:
        op_str = self.VALID_OPERATORS[self.operator]
        if isinstance(self.left, CompositeParameter):
            left_repr = self.left._bare_repr()
        elif isinstance(self.left, Parameter):
            left_argspec = self.left._get_argspec()
            left_repr = function_repr(self.left.func, left_argspec)
        else:
            left_repr = str(self.left)

        if isinstance(self.right, CompositeParameter):
            right_repr = self.right._bare_repr()
        elif isinstance(self.right, Parameter):
            right_argspec = self.right._get_argspec()
            right_repr = function_repr(self.right.func, right_argspec)
        else:
            right_repr = str(self.right)

        return f"({left_repr} {op_str} {right_repr})"

    def __eq__(self, other) -> bool:
        if other is self:
            return True

        if not isinstance(other, type(self)):
            return False

        return (
            self.left == other.left
            and self.right == other.right
            and self.operator is other.operator
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}<{self._bare_repr()}>"

    def __getstate__(self):
        state = self.__dict__.copy()
        state["left"] = cloudpickle.dumps(state["left"])
        state["right"] = cloudpickle.dumps(state["right"])
        return state

    def __setstate__(self, state):
        state["left"] = cloudpickle.loads(state["left"])
        state["right"] = cloudpickle.loads(state["right"])
        self.__dict__.update(state)


class Constant(Parameter):
    """A Parameter whose value doesn't depend on position."""

    def __init__(self, value: Union[int, float, complex], dimensions: int = 2):
        if dimensions not in (2, 3):
            raise ValueError(f"Dimensions must be 2 or 3, got {dimensions}.")
        if dimensions == 2:

            def constant(x, y, value=0):
                return value * np.ones_like(x)

        else:

            def constant(x, y, z, value=0):
                return value * np.ones_like(x)

        super().__init__(constant, value=value)



from typing import Optional, Tuple

import numpy as np


def rotation_matrix(angle_radians: float) -> np.ndarray:
    """Returns a 2D rotation matrix."""
    c = np.cos(angle_radians)
    s = np.sin(angle_radians)
    return np.array([[c, -s], [s, c]])


def rotate(coords: np.ndarray, angle_degrees: float) -> np.ndarray:
    """Rotates an array of (x, y) coordinates counterclockwise by
    the specified angle.

    Args:
        coords: Shape (n, 2) array of (x, y) coordinates.
        angle_degrees: The angle by which to rotate the coordinates.

    Returns:
        Shape (n, 2) array of rotated coordinates (x', y')
    """
    coords = np.asarray(coords)
    assert coords.ndim == 2
    assert coords.shape[1] == 2
    R = rotation_matrix(np.radians(angle_degrees))
    return (R @ coords.T).T


def ellipse(
    a: float,
    b: float,
    points: int = 100,
    center: Tuple[float, float] = (0, 0),
    angle: float = 0,
):
    """Returns the coordinates for an ellipse with major axis a and semimajor axis b,
    rotated by the specified angle about (0, 0), then translated to the specified center.

    Args:
        a: Major axis length
        b: Semi-major axis length
        points: Number of points in the circle
        center: Coordinates of the center of the circle
        angle: Angle (in degrees) by which to rotate counterclockwise about (0, 0)
            **after** translating to the specified center.

    Returns:
        A shape ``(points, 2)`` array of (x, y) coordinates
    """
    x0, y0 = center
    theta = np.linspace(0, 2 * np.pi, points, endpoint=False)
    xs = a * np.cos(theta)
    ys = b * np.sin(theta)
    coords = np.array([xs, ys]).T + np.array([[x0, y0]])
    if angle:
        coords = rotate(coords, angle)
    return coords


def circle(
    radius: float, points: int = 100, center: Tuple[float, float] = (0, 0)
) -> np.ndarray:
    """Returns the coordinates for a circle with a given radius, centered at the
    specified center.

    Args:
        radius: Radius of the circle
        points: Number of points in the circle
        center: Coordinates of the center of the circle

    Returns:
        A shape ``(points, 2)`` array of (x, y) coordinates
    """
    return ellipse(
        radius,
        radius,
        points=points,
        center=center,
        angle=0,
    )


def box(
    width: float,
    height: Optional[float] = None,
    points: int = 101,
    center: Tuple[float, float] = (0, 0),
    angle: float = 0,
) -> np.ndarray:
    """Returns the coordinates for a rectangle with a given width and height,
    centered at the specified center.

    Args:
        width: Width of the rectangle (in the x direction).
        height: Height of the rectangle (in the y direction). If None is given,
            then height is set to width and the function returns a square.
        points: The target number of points making up the box. The actual number of
            points may be slightly different than this value.
        center: Coordinates of the center of the rectangle.
        angle: Angle (in degrees) by which to rotate counterclockwise about (0, 0)
            **after** translating to the specified center.

    Returns:
        A shape ``(m, 2)`` or array of (x, y) coordinates.
    """
    width = abs(width)
    if height is None:
        height = width
    height = abs(height)
    x0, y0 = center
    perimeter = 2 * (width + height)
    x_points = round(points * width / perimeter)
    y_points = round(points * height / perimeter)

    xs = np.concatenate(
        [
            width / 2 * np.ones(y_points),
            np.linspace(width / 2, -width / 2, x_points),
            -width / 2 * np.ones(y_points),
            np.linspace(-width / 2, width / 2, x_points),
        ]
    )
    ys = np.concatenate(
        [
            np.linspace(-height / 2, height / 2, y_points),
            height / 2 * np.ones(x_points),
            np.linspace(height / 2, -height / 2, y_points),
            -height / 2 * np.ones(x_points),
        ]
    )
    coords = np.array([xs, ys]).T + np.array([[x0, y0]])
    if angle:
        coords = rotate(coords, angle)
    return coords


def close_curve(points: np.ndarray) -> np.ndarray:
    """Close a curve (making the start point equal to the end point),
    if it is not already closed.

    Args:
        points: Shape ``(m, n)`` array of ``m`` coordinates in ``n`` dimensions.

    Returns:
        ``points`` with the first point appended to the end if the start point
        was not already equal to the end point.
    """
    if not np.allclose(points[0], points[-1]):
        points = np.concatenate([points, points[:1]], axis=0)
    return points


def ensure_unique(coords: np.ndarray) -> np.ndarray:
    # Coords is a shape (n, 2) array of vertex coordinates.
    coords = np.asarray(coords)
    # Remove duplicate coordinates, otherwise triangle.build() will segfault.
    # By default, np.unique() does not preserve order, so we have to remove
    # duplicates this way:
    _, ix = np.unique(coords, return_index=True, axis=0)
    coords = coords[np.sort(ix)]
    return coords


def unit_vector(vector: np.ndarray) -> np.ndarray:
    """Normalizes ``vector``."""
    return vector / np.linalg.norm(vector, axis=-1)[:, np.newaxis]


def path_vectors(path: np.ndarray) -> Tuple[float, np.ndarray]:
    """Computes the total length and the unit normals for a path.

    Args:
        path: Shape ``(n, 2)`` array of coordinates representing a continuous path.

    Returns:
        A shape ``(n-1, 2)`` array of edge lengths and a shape ``(n-1, 2)`` array of
        unit vectors normal to each edge in the path.
    """
    dr = np.diff(path, axis=0)
    normals = np.cross(dr, [0, 0, 1])
    unit_normals = unit_vector(normals)
    edge_lengths = np.linalg.norm(dr, axis=1)
    return edge_lengths, unit_normals[:, :2]



import matplotlib.pyplot as plt
import numpy as np
import pytest

import tdgl

ys = np.linspace(-5, 5, 501)
xs = np.ones_like(ys)
cross_section_coord_params = [
    None,
    np.array([0 * xs, ys]).T,
    np.array([8 * xs, ys]).T,
    [
        np.array([0 * xs, ys]).T,
        np.array([8 * xs, ys]).T,
    ],
]


def test_plot_order_parameter(transport_device_solution: tdgl.Solution):
    fig, axes = transport_device_solution.plot_order_parameter()
    assert isinstance(fig, plt.Figure)
    assert len(axes) == 2
    plt.close(fig)


@pytest.mark.parametrize("vmin, vmax", [(None, None), (0, 5), (0, None)])
@pytest.mark.parametrize("streamplot", [False, True])
@pytest.mark.parametrize("cross_section_coords", cross_section_coord_params)
@pytest.mark.parametrize("units", [None, "uA/um", "mA/um"])
@pytest.mark.parametrize("dataset", [None, "supercurrent", "normal_current"])
@pytest.mark.parametrize("colorbar", [False, True])
def test_plot_currents(
    transport_device_solution: tdgl.Solution,
    vmin,
    vmax,
    streamplot,
    cross_section_coords,
    units,
    dataset,
    colorbar,
):
    solution = transport_device_solution
    kwargs = dict(
        vmin=vmin,
        vmax=vmax,
        streamplot=streamplot,
        cross_section_coords=cross_section_coords,
        units=units,
        dataset=dataset,
        colorbar=colorbar,
    )

    if vmax is None and vmin is not None:
        with pytest.raises(ValueError):
            _ = solution.plot_currents(**kwargs)
    else:
        fig, ax = solution.plot_currents(**kwargs)
    plt.close("all")


@pytest.mark.parametrize("vector", [False, True])
@pytest.mark.parametrize(
    "positions, zs",
    [
        (np.random.rand(200).reshape((-1, 2)), np.random.rand(100)),
        (np.random.rand(200).reshape((-1, 2)), 1),
        (np.random.rand(300).reshape((-1, 3)), None),
    ],
)
@pytest.mark.parametrize("units", [None, "mT"])
@pytest.mark.parametrize("auto_range_cutoff", [None, 1])
@pytest.mark.parametrize("cross_section_coords", cross_section_coord_params)
def test_plot_field_at_positions(
    transport_device_solution,
    positions,
    zs,
    units,
    cross_section_coords,
    auto_range_cutoff,
    vector,
):
    solution = transport_device_solution
    fig, ax = solution.plot_field_at_positions(
        positions,
        zs=zs,
        vector=vector,
        grid_shape=(50, 50),
        units=units,
        cross_section_coords=cross_section_coords,
        auto_range_cutoff=auto_range_cutoff,
        share_color_scale=True,
    )
    plt.close(fig)


@pytest.mark.parametrize("vmin, vmax", [(None, None), (-5, 5), (-5, None)])
@pytest.mark.parametrize("auto_range_cutoff", [None, 1, (5, 95)])
@pytest.mark.parametrize("units", [None, "uA/um**2"])
@pytest.mark.parametrize("symmetric_color_scale", [False, True])
def test_plot_vorticity(
    transport_device_solution,
    vmin,
    vmax,
    auto_range_cutoff,
    units,
    symmetric_color_scale,
):
    solution = transport_device_solution
    kwargs = dict(
        vmin=vmin,
        vmax=vmax,
        units=units,
        auto_range_cutoff=auto_range_cutoff,
        symmetric_color_scale=symmetric_color_scale,
    )

    if vmax is None and vmin is not None:
        with pytest.raises(ValueError):
            _ = solution.plot_vorticity(**kwargs)
    else:
        fig, ax = solution.plot_vorticity(**kwargs)
    plt.close("all")


@pytest.mark.parametrize("vmin, vmax", [(None, None), (0, 5), (0, None)])
@pytest.mark.parametrize("auto_range_cutoff", [None, 1, (5, 95)])
def test_plot_scalar_potential(
    transport_device_solution,
    vmin,
    vmax,
    auto_range_cutoff,
):
    solution = transport_device_solution
    kwargs = dict(
        vmin=vmin,
        vmax=vmax,
        auto_range_cutoff=auto_range_cutoff,
    )

    if vmax is None and vmin is not None:
        with pytest.raises(ValueError):
            _ = solution.plot_scalar_potential(**kwargs)
    else:
        fig, ax = solution.plot_scalar_potential(**kwargs)
    plt.close("all")



import logging
from typing import Any, Dict, Sequence, Union

import h5py
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.ticker import FuncFormatter

from ..finite_volume.mesh import Mesh
from ..solution.data import get_data_range
from .common import DEFAULT_QUANTITIES, PLOT_DEFAULTS, Quantity, auto_grid
from .io import get_plot_data, get_state_string


class InteractivePlot:
    def __init__(
        self,
        input_file: str,
        logger: logging.Logger = None,
    ):

        self.input_file = input_file
        self.frame = 0
        self.quantity = Quantity.ORDER_PARAMETER
        self.logger = logger or logging.getLogger()

    def show(self):
        with h5py.File(self.input_file, "r", libver="latest") as h5file:
            if "mesh" in h5file:
                mesh = Mesh.from_hdf5(h5file["mesh"])
            else:
                mesh = Mesh.from_hdf5(h5file["solution/device/mesh"])

            # Get the ranges for the frame
            min_frame, max_frame = get_data_range(h5file)

            def on_keypress(event):
                if event.key == "right":
                    self.frame = np.minimum(self.frame + 1, max_frame)

                elif event.key == "left":
                    self.frame = np.maximum(self.frame - 1, min_frame)

                if event.key == "shift+right":
                    self.frame = np.minimum(self.frame + 10, max_frame)

                elif event.key == "shift+left":
                    self.frame = np.maximum(self.frame - 10, min_frame)

                elif event.key == "up":
                    self.frame = np.minimum(self.frame + 100, max_frame)

                elif event.key == "down":
                    self.frame = np.maximum(self.frame - 100, min_frame)

                elif event.key == "shift+up":
                    self.frame = np.minimum(self.frame + 1000, max_frame)

                elif event.key == "shift+down":
                    self.frame = np.maximum(self.frame - 1000, min_frame)

                elif event.key == "home":
                    self.frame = min_frame

                elif event.key == "end":
                    self.frame = max_frame

                elif event.key == "1":
                    self.quantity = Quantity.ORDER_PARAMETER

                elif event.key == "2":
                    self.quantity = Quantity.PHASE

                elif event.key == "3":
                    self.quantity = Quantity.SUPERCURRENT

                elif event.key == "4":
                    self.quantity = Quantity.NORMAL_CURRENT

                elif event.key == "5":
                    self.quantity = Quantity.SCALAR_POTENTIAL

                elif event.key == "6":
                    self.quantity = Quantity.APPLIED_VECTOR_POTENTIAL

                elif event.key == "7":
                    self.quantity = Quantity.INDUCED_VECTOR_POTENTIAL

                elif event.key == "8":
                    self.quantity = Quantity.EPSILON

                elif event.key == "9":
                    self.quantity = Quantity.VORTICITY

                draw()

            def draw():
                value, direction, limits = get_plot_data(
                    h5file, mesh, self.quantity, self.frame
                )
                state = get_state_string(h5file, self.frame, max_frame)

                fig.suptitle(f"{self.quantity.value}\n{state}")
                triplot.set_array(value)
                triplot.set_clim(*limits)
                triplot.set_cmap(PLOT_DEFAULTS[self.quantity].cmap)
                cbar.set_label(PLOT_DEFAULTS[self.quantity].clabel)
                fig.canvas.draw()

            x, y = mesh.sites.T
            # Temp data to use in plots
            temp_value = np.ones_like(x)
            temp_value[0] = 0
            temp_value[1] = 0.5

            fig, ax = plt.subplots()
            fig.subplots_adjust(top=0.8)
            fig.canvas.mpl_connect("key_press_event", on_keypress)
            triplot = ax.tripcolor(
                x, y, temp_value, triangles=mesh.elements, shading="gouraud"
            )
            cbar = fig.colorbar(triplot)
            ax.set_aspect("equal")
            draw()
            plt.show()


class MultiInteractivePlot:
    def __init__(
        self,
        input_file: str,
        quantities: Sequence[str] = DEFAULT_QUANTITIES,
        max_cols: int = 4,
        logger: logging.Logger = None,
        figure_kwargs: Union[Dict[str, Any], None] = None,
    ):
        self.input_file = input_file
        self.frame = 0
        if quantities is None:
            quantities = Quantity.get_keys()
        self.quantities = [Quantity.from_key(name) for name in quantities]
        self.num_plots = len(quantities)
        self.max_cols = max_cols
        self.logger = logger if logger is not None else logging.getLogger()
        self.figure_kwargs = figure_kwargs or dict()
        self.figure_kwargs.setdefault("constrained_layout", True)
        default_figsize = (
            3.25 * min(self.max_cols, self.num_plots),
            3 * max(1, self.num_plots // self.max_cols),
        )
        self.figure_kwargs.setdefault("figsize", default_figsize)

    def show(self):
        with h5py.File(self.input_file, "r", libver="latest") as h5file:
            if "mesh" in h5file:
                mesh = Mesh.from_hdf5(h5file["mesh"])
            else:
                mesh = Mesh.from_hdf5(h5file["solution/device/mesh"])

            min_frame, max_frame = get_data_range(h5file)

            def on_keypress(event):
                if event.key == "right":
                    self.frame = np.minimum(self.frame + 1, max_frame)

                elif event.key == "left":
                    self.frame = np.maximum(self.frame - 1, min_frame)

                if event.key == "shift+right":
                    self.frame = np.minimum(self.frame + 10, max_frame)

                elif event.key == "shift+left":
                    self.frame = np.maximum(self.frame - 10, min_frame)

                elif event.key == "up":
                    self.frame = np.minimum(self.frame + 100, max_frame)

                elif event.key == "down":
                    self.frame = np.maximum(self.frame - 100, min_frame)

                elif event.key == "shift+up":
                    self.frame = np.minimum(self.frame + 1000, max_frame)

                elif event.key == "shift+down":
                    self.frame = np.maximum(self.frame - 1000, min_frame)

                elif event.key == "home":
                    self.frame = min_frame

                elif event.key == "end":
                    self.frame = max_frame

                draw()

            vmins = [+np.inf for _ in self.quantities]
            vmaxs = [-np.inf for _ in self.quantities]

            def draw():
                state = get_state_string(h5file, self.frame, max_frame)
                fig.suptitle(state)
                for i, (quantity, collection) in enumerate(
                    zip(self.quantities, collections)
                ):
                    value, direction, limits = get_plot_data(
                        h5file, mesh, quantity, self.frame
                    )
                    collection.set_array(value)
                    vmins[i] = min(vmins[i], limits[0])
                    vmaxs[i] = max(vmaxs[i], limits[1])
                    collection.set_clim(vmins[i], vmaxs[i])
                fig.canvas.draw()

            x, y = mesh.sites.T
            # Temp data to use in plots
            temp_value = np.ones_like(x)
            temp_value[0] = 0
            temp_value[1] = 0.5

            fig, axes = auto_grid(
                self.num_plots, max_cols=self.max_cols, **self.figure_kwargs
            )
            fig.canvas.mpl_connect("key_press_event", on_keypress)

            collections = []
            for quantity, ax in zip(self.quantities, axes.flat):
                opts = PLOT_DEFAULTS[quantity]
                collection = ax.tripcolor(
                    x,
                    y,
                    temp_value,
                    triangles=mesh.elements,
                    shading="gouraud",
                    cmap=opts.cmap,
                )
                cbar = fig.colorbar(
                    collection, ax=ax, format=FuncFormatter("{:.2f}".format)
                )
                cbar.set_label(opts.clabel)
                ax.set_aspect("equal")
                ax.set_title(quantity.value)
                collections.append(collection)

            draw()
            plt.show()



import warnings
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Sequence, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np


class Quantity(Enum):
    ORDER_PARAMETER = "Order parameter"
    PHASE = "Phase"
    SUPERCURRENT = "Supercurrent density"
    NORMAL_CURRENT = "Normal current density"
    VORTICITY = "Vorticity"
    SCALAR_POTENTIAL = "Scalar potential"
    APPLIED_VECTOR_POTENTIAL = "Applied vector potential"
    INDUCED_VECTOR_POTENTIAL = "Induced vector potential"
    EPSILON = "Epsilon"

    @classmethod
    def get_keys(cls) -> Sequence[str]:
        return list(item.name for item in Quantity)

    @classmethod
    def from_key(cls, key: str) -> "Quantity":
        return Quantity[key.upper()]


colormaps = {
    Quantity.ORDER_PARAMETER: "viridis",
    Quantity.PHASE: "twilight_shifted",
    Quantity.SUPERCURRENT: "inferno",
    Quantity.NORMAL_CURRENT: "inferno",
    Quantity.SCALAR_POTENTIAL: "magma",
    Quantity.APPLIED_VECTOR_POTENTIAL: "cividis",
    Quantity.INDUCED_VECTOR_POTENTIAL: "cividis",
    Quantity.EPSILON: "viridis",
    Quantity.VORTICITY: "coolwarm",
}


@dataclass
class PlotDefault:
    cmap: str
    clabel: str
    xlabel: str = "$x/\\xi$"
    ylabel: str = "$y/\\xi$"


PLOT_DEFAULTS = {
    Quantity.ORDER_PARAMETER: PlotDefault(cmap="viridis", clabel="$|\\psi|$"),
    Quantity.PHASE: PlotDefault(
        cmap="twilight_shifted",
        clabel="$\\arg(\\psi)/\\pi$",
    ),
    Quantity.SUPERCURRENT: PlotDefault(cmap="inferno", clabel="$|\\vec{{J}}_s|/J_0$"),
    Quantity.NORMAL_CURRENT: PlotDefault(cmap="inferno", clabel="$|\\vec{{J}}_n|/J_0$"),
    Quantity.SCALAR_POTENTIAL: PlotDefault(cmap="magma", clabel="$\\mu/v_0$"),
    Quantity.APPLIED_VECTOR_POTENTIAL: PlotDefault(
        cmap="cividis", clabel="$a_\\mathrm{{applied}}/(\\xi B_{{c2}})$"
    ),
    Quantity.INDUCED_VECTOR_POTENTIAL: PlotDefault(
        cmap="cividis", clabel="$a_\\mathrm{{induced}}/(\\xi B_{{c2}})$"
    ),
    Quantity.EPSILON: PlotDefault(cmap="viridis", clabel="$\\epsilon$"),
    Quantity.VORTICITY: PlotDefault(
        cmap="coolwarm", clabel="$(\\vec{{\\nabla}}\\times\\vec{{J}})\\cdot\\hat{{z}}$"
    ),
}

DEFAULT_QUANTITIES = (
    "order_parameter",
    "phase",
    "supercurrent",
    "normal_current",
)


def auto_grid(
    num_plots: int,
    max_cols: int = 3,
    delaxes: bool = True,
    **kwargs,
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Creates a grid of at least ``num_plots`` subplots
    with at most ``max_cols`` columns.

    Additional keyword arguments are passed to ``plt.subplots()``.

    Args:
        num_plots: Total number of plots that will be populated.
        max_cols: Maximum number of columns in the grid.
        delaxes: Whether to remove unused axes.

    Returns:
        matplotlib figure and axes
    """
    ncols = min(max_cols, num_plots)
    nrows = int(np.ceil(num_plots / ncols))
    fig, axes = plt.subplots(nrows, ncols, **kwargs)
    if not isinstance(axes, (list, np.ndarray)):
        axes = np.array([axes])
    axes = np.asarray(axes)
    if delaxes:
        flat_axes = list(axes.flat)
        for ax in flat_axes[num_plots:]:
            fig.delaxes(ax)
    return fig, axes


@contextmanager
def non_gui_backend():
    """A contextmanager that temporarily uses a non-GUI backend for matplotlib."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", category=UserWarning, message="Matplotlib is currently using agg"
        )
        try:
            old_backend = mpl.get_backend()
            mpl.use("Agg")
            yield
        finally:
            mpl.use(old_backend)


def auto_range_iqr(
    data_array: np.ndarray,
    cutoff_percentile: Union[float, Tuple[float, float]] = 1,
) -> Tuple[float, float]:
    """Get the min and max range of the provided array that excludes outliers
    following the IQR rule.

    This function computes the inter-quartile-range (IQR), defined by Q3-Q1,
    i.e. the percentiles for 75 and 25 percent of the distribution. The region
    without outliers is defined by [Q1-1.5*IQR, Q3+1.5*IQR].
    Taken from `qcodes <https://github.com/QCoDeS/Qcodes/blob/
    6c8f7202f6b6fca4884bfc0f6e1e9a6564628d75/qcodes/utils/plotting.py#L28-L76>`_.

    Args:
        data_array: Array of arbitrary dimension containing the
            statistical data.
        cutoff_percentile: Percentile of data that may maximally be
            clipped on both sides of the distribution. If given a
            tuple (a, b) the percentile limits will be a and 100-b.

    Returns:
        vmin, vmax
    """
    if isinstance(cutoff_percentile, tuple):
        bottom, top = cutoff_percentile
    else:
        bottom = cutoff_percentile
        top = 100 - bottom
    z = data_array.flatten()
    zmax = np.nanmax(z)
    zmin = np.nanmin(z)
    zrange = zmax - zmin
    pmin, q3, q1, pmax = np.nanpercentile(z, [bottom, 75, 25, top])
    iqr = q3 - q1
    # handle corner case of all data zero, such that IQR is zero
    # to counter numerical artifacts do not test IQR == 0, but IQR on its
    # natural scale (zrange) to be smaller than some very small number.
    # also test for zrange to be 0.0 to avoid division by 0.
    if zrange == 0.0 or iqr / zrange < 1e-8:
        vmin = zmin
        vmax = zmax
    else:
        vmin = max(q1 - 1.5 * iqr, zmin)
        vmax = min(q3 + 1.5 * iqr, zmax)
        vmin = min(vmin, pmin)
        vmax = max(vmax, pmax)
    return vmin, vmax



from typing import Optional, Sequence, Union

import numpy as np
import pint
from scipy import spatial, special
from scipy.constants import mu_0

from .finite_volume.mesh import Mesh

ureg = pint.UnitRegistry()


def convert_field(
    value: Union[np.ndarray, float, str, pint.Quantity],
    new_units: Union[str, pint.Unit],
    old_units: Optional[Union[str, pint.Unit]] = None,
    ureg: Optional[pint.UnitRegistry] = None,
    with_units: bool = True,
) -> Union[pint.Quantity, np.ndarray, float]:
    """Converts a value between different field units, either magnetic field H
    [current] / [length] or flux density B = mu0 * H [mass] / ([curret] [time]^2)).

    Args:
        value: The value to convert. It can either be a numpy array (no units),
            a float (no units), a string like "1 uA/um", or a scalar or array
            ``pint.Quantity``. If value is not a string wiht units or a ``pint.Quantity``,
            then old_units must specify the units of the float or array.
        new_units: The units to convert to.
        old_units: The old units of ``value``. This argument is required if ``value``
            is not a string with units or a ``pint.Quantity``.
        ureg: The ``pint.UnitRegistry`` to use for conversion. If None is given,
            a new instance is created.
        with_units: Whether to return a ``pint.Quantity`` with units attached.

    Returns:
        The converted value, either a pint.Quantity (scalar or array with units),
        or an array or float without units, depending on the ``with_units`` argument.
    """
    if ureg is None:
        ureg = pint.UnitRegistry()
    if isinstance(value, str):
        value = ureg(value)
    if isinstance(value, pint.Quantity):
        old_units = value.units
    if old_units is None:
        raise ValueError(
            "Old units must be specified if value is not a string or pint.Quantity."
        )
    if isinstance(old_units, str):
        old_units = ureg(old_units)
    if isinstance(new_units, str):
        new_units = ureg(new_units)
    if not isinstance(value, pint.Quantity):
        value = value * old_units
    if new_units.dimensionality == old_units.dimensionality:
        value = value.to(new_units)
    elif "[length]" in old_units.dimensionality:
        # value is H in units with dimensionality [current] / [length]
        # and we want B = mu0 * H
        value = (value * ureg("mu0")).to(new_units)
    else:
        # value is B = mu0 * H in units with dimensionality
        # [mass] / ([current] [time]^2) and we want H = B / mu0
        value = (value / ureg("mu0")).to(new_units)
    if not with_units:
        value = value.magnitude
    return value


def biot_savart(
    eval_positions: np.ndarray,
    *,
    current_positions: np.ndarray,
    current_vectors: np.ndarray,
    currents: np.ndarray,
) -> np.ndarray:
    """Calculates the vector magnetic field [Bx, By, Bx] at ``eval_positions``
    due to a discrete set of 1D current elements.

    Input units are meters and Amperes, output units are Tesla.

    Args:
        eval_positions: Shape (n, 3) array of (x, y, z) positions at which to
            evaluate the field.
        current_positions: Shape (m, 3) array of (x, y, z) positions for the
            current elements.
        current_vectors: Shape (m, 3) array of (dx, dy, dy) distance vectors
            indicating the direction and length of the current elements.
        currents: Shape (m, ) or (m, 1) array of current magnitudes for each
            current element.

    Returns:
        Shape (n, 3) pint.Quantity array of the vector magnetic field
        at ``eval_positions``.
    """
    eval_positions = np.atleast_2d(eval_positions)
    current_positions = np.atleast_2d(current_positions)
    current_vectors = np.atleast_2d(current_vectors)
    currents = np.atleast_2d(currents)
    assert eval_positions.shape[-1] == 3
    assert current_positions.shape[-1] == 3
    assert current_vectors.shape[-1] == 3
    assert currents.shape[-1] == 1

    dx = np.subtract.outer(eval_positions[:, 0], current_positions[:, 0])
    dy = np.subtract.outer(eval_positions[:, 1], current_positions[:, 1])
    dz = np.subtract.outer(eval_positions[:, 2], current_positions[:, 2])
    rprime = np.stack([dx, dy, dz], axis=-1)
    denom = (np.linalg.norm(rprime, axis=-1) ** 3)[:, :, np.newaxis]
    integrand = np.cross(current_vectors, rprime) / denom
    integral = np.einsum("jk,ijk -> ik", currents, integrand)
    return mu_0 / (4 * np.pi) * integral * ureg("tesla")


def biot_savart_2d(
    x: Union[float, np.ndarray],
    y: Union[float, np.ndarray],
    z: Union[float, np.ndarray],
    *,
    positions: np.ndarray,
    current_densities: np.ndarray,
    z0: float = 0,
    areas: Optional[np.ndarray] = None,
    length_units: str = "um",
    current_units: str = "uA",
    vector: bool = True,
) -> np.ndarray:
    """Returns the magnetic field (in tesla) from a sheet of current located at
    vertical positon ``z0`` (in units of ``length_units``). The current is
    parameterized by a set of ``current_densities`` (in units of
    ``current_units / length_units``) and x-y ``positions`` (in units of
    ``length_units``), and the field is evaluated at coordinates ``(x, y, z)``.

    .. math::

        \\mu_0H_x(\\vec{r}) &= \\frac{\\mu_0}{4\\pi}\\int_S
        \\frac{J_y(\\vec{r}')(\\vec{r}-\\vec{r}')\\cdot\\hat{z}}
        {|\\vec{r}-\\vec{r}'|^3}\\,\\mathrm{d}^2r'\\\\
        \\mu_0H_y(\\vec{r}) &= \\frac{\\mu_0}{4\\pi}\\int_S
        -\\frac{J_x(\\vec{r}')(\\vec{r}-\\vec{r}')\\cdot\\hat{z}}
        {|\\vec{r}-\\vec{r}'|^3}\\,\\mathrm{d}^2r'\\\\
        \\mu_0H_z(\\vec{r}) &= \\frac{\\mu_0}{4\\pi}\\int_S
        \\frac{J_x(\\vec{r}')(\\vec{r}-\\vec{r}')\\cdot\\hat{y}
        - J_y(\\vec{r}')(\\vec{r}-\\vec{r}')\\cdot\\hat{x}}
        {|\\vec{r}-\\vec{r}'|^3}\\,\\mathrm{d}^2r'


    where :math:`\\vec{r}=(x, y, z)` and :math:`\\vec{r}'=(x', y', z_0)`.

    Args:
        x: x-coordinate(s) at which to evaluate the field.
            Either a scalar or vector with shape ``(n, )``.
        y: y-coordinate(s) at which to evaluate the field.
            Either a scalar or vector with shape ``(n, )``.
        z: z-coordinate(s) at which to evaluate the field. Either a scalar
            or vector with shape ``(n, )``.
        positions: Coordinates ``(x0, y0)`` of the current sheet,
            shape ``(m, 2)``.
        current_densities: 2D current density ``(Jx, Jy)``, shape``(m, 2)``.
        z0: Vertical (z) position of the current sheet.
        areas: Vertex areas for ``positions`` in units of ``length_units**2``. If None,
            the ``positions`` are triangulated to calculate vertex areas.
        length_units: The units for all coordinates.
        current_units: The units for current values. The ``current_densities`` are
            assumed to be in units of ``current_units / length_units``.
        vector: Return the full vector magnetic field (shape ``(n, 3)``) rather
            than just the z-component (shape ``(n, )``).

    Returns:
        Magnetic field in tesla evaluated at ``(x, y, z)``. If ``vector`` is True,
        returns the vector magnetic field :math:`\\mu_0\\vec{H}` (shape ``(n, 3)``).
        Otherwise, returns the the :math:`z`-component, :math:`\\mu_0H_z`
        (shape ``(n,)``).
    """
    # Convert everything to base units: meters and amps / meter.
    to_meter = ureg(length_units).to("m").magnitude
    to_amp_per_meter = ureg(f"{current_units} / {length_units}").to("A / m").magnitude
    x, y, z = np.atleast_1d(x, y, z)
    if z.shape[0] == 1:
        z = z * np.ones_like(x)
    x = x * to_meter
    y = y * to_meter
    z = z * to_meter
    positions, current_densities = np.atleast_2d(positions, current_densities)
    positions = positions * to_meter
    z0 = z0 * to_meter
    current_densities = current_densities * to_amp_per_meter
    # Calculate the pairwise distance between the current sheet and evaluation
    # points for each axis.
    x0, y0 = positions[:, 0], positions[:, 1]
    Jx, Jy = current_densities[:, 0], current_densities[:, 1]
    dx = np.subtract.outer(x, x0)
    dy = np.subtract.outer(y, y0)
    dz = np.subtract.outer(z, z0 * np.ones_like(x0))
    if areas is None:
        # Triangulate the current sheet to assign an effective area to each vertex.
        triangles = spatial.Delaunay(positions).simplices
        mesh = Mesh.from_triangulation(positions[:, 0], positions[:, 1], triangles)
        areas = mesh.areas
    else:
        areas = areas * to_meter**2
    # Evaluate the Biot-Savart integral.
    pref = (mu_0 / (4 * np.pi)) * areas * (dx**2 + dy**2 + dz**2) ** (-3 / 2)
    Jx_dy = np.einsum("ij, ij, j -> i", pref, dy, Jx)
    Jy_dx = np.einsum("ij, ij, j -> i", pref, dx, Jy)
    Bz = Jx_dy - Jy_dx
    if not vector:
        return Bz
    Jy_dz = np.einsum("ij, ij, j -> i", pref, dz, Jy)
    Jx_dz = np.einsum("ij, ij, j -> i", pref, dz, Jx)
    Bx = Jy_dz
    By = -Jx_dz
    return np.array([Bx, By, Bz]).T * ureg("tesla")


def current_loop_vector_potential(
    positions: np.ndarray,
    *,
    loop_center: Sequence[float] = (0, 0, 0),
    loop_radius: float = 1,
    current: float = 1,
    length_units: str = "um",
    current_units: str = "uA",
):
    """Calculates the magnetic vector potential [Ax, Ay] at ``positions``
    due to a 1D current loop.

    Args:
        positions: Shape (n, 3) array of (x, y, z) positions at which to
            evaluate the vector potential.
        loop_center: (x, y, z) coordinates of the current loop center.
        loop_radius: radius of the current loop.
        current: Magnitude of the current flowing in the loop.
        length_units: A string specifying the length units.
        current_units: A string specifying the current units.

    Returns:
        Shape (n, 3) array of the vector potential [Ax, Ay, Az] at ``positions``.
    """
    to_meter = ureg(length_units).to("m").magnitude
    to_amp = ureg(current_units).to("A").magnitude
    # http://www.physics.usu.edu/Wheeler/EMarchive/Jch5Notes.pdf
    positions = np.atleast_2d(positions) * to_meter
    loop_center = np.atleast_2d(loop_center) * to_meter
    a = loop_radius * to_meter
    current = current * to_amp
    positions = positions - loop_center
    # # This is a pint-friendly vector norm.
    # rs = np.sqrt(np.sum(np.square(positions), axis=1))
    rs = np.linalg.norm(positions, axis=1)
    thetas = np.arccos(positions[:, 2] / rs)
    sin_thetas = np.sin(thetas)
    # m == k**2, see docs for scipy.special.ellipk
    denom = rs**2 + a**2 + 2 * a * rs * sin_thetas
    m = 4 * a * rs * sin_thetas / denom
    K = special.ellipk(m)
    E = special.ellipe(m)
    mag = -mu_0 * current * a / (np.pi * m) * (((m - 2) * K + 2 * E)) / np.sqrt(denom)
    # \vec{A} is directed along the azimuthal direction,
    # so here we generate the azimuthal unit vector.
    # Azimuthal angle + pi / 2 to get azimuthal direction.
    phis = np.arctan2(positions[:, 1], positions[:, 0]) + np.pi / 2
    direc = np.array([np.cos(phis), np.sin(phis), np.zeros_like(phis)]).T
    return mag[:, np.newaxis] * direc * ureg("T * m")


def current_loop_field(
    positions: np.ndarray,
    *,
    loop_center: Sequence[float] = (0, 0, 0),
    loop_radius: float = 1e-6,
    current: float = 1e-3,
    num_segments: int = 101,
    length_units: str = "um",
    current_units: str = "uA",
):
    """Calculates the vector magnetic field [Bx, By, Bz] at ``positions``
    due to a 1D current loop, in units of tesla.

    Args:
        positions: Shape (n, 3) array of (x, y, z) positions at which to
            evaluate the vector potential.
        loop_center: (x, y, z) coordinates of the current loop center.
        loop_radius: radius of the current loop.
        current: Magnitude of the current flowing in the loop.
        num_segments: Number of current elements used to model the loop.
        length_units: A string specifying the length units.
        current_units: A string specifying the current units.

    Returns:
        Shape (n, 3) array of the magnetic field [Bx, By, Bz] at ``positions``.
    """
    to_meter = ureg(length_units).to("m").magnitude
    to_amp = ureg(current_units).to("A").magnitude
    positions = np.atleast_2d(positions) * to_meter
    loop_center = np.atleast_2d(loop_center) * to_meter
    loop_radius = loop_radius * to_meter
    current = current * to_amp
    # Create loop positions
    thetas = np.linspace(0, 2 * np.pi, num_segments)
    circ = np.array([np.cos(thetas), np.sin(thetas), np.zeros_like(thetas)]).T
    loop = loop_radius * circ + loop_center
    dloop = np.diff(loop, axis=0)
    loop = loop[:-1]
    return biot_savart(
        positions,
        current_positions=loop,
        current_vectors=dloop,
        currents=current,
    ).to("tesla")


def uniform_Bz_vector_potential(
    positions: np.ndarray,
    Bz: Union[float, str, pint.Quantity],
) -> np.ndarray:
    """Calculates the magnetic vector potential [Ax, Ay, Az] at ``positions``
    due uniform magnetic field along the z-axis with strength ``Bz``.

    Args:
        positions: Shape (n, 3) array of (x, y, z) positions in meters at which to
            evaluate the vector potential.
        Bz: The strength of the uniform field, as a pint-parseable string,
            a pint.Quantity, or a float with units of Tesla.

    Returns:
        Shape (n, 3) array of the vector potential [Ax, Ay, Az] at ``positions``
        in units of Tesla * meter.
    """
    assert isinstance(Bz, (float, str, pint.Quantity)), type(Bz)
    positions = np.atleast_2d(positions)
    assert positions.shape[1] == 3, positions.shape
    if not isinstance(positions, pint.Quantity):
        positions = positions * ureg("meter")
    if isinstance(Bz, str):
        Bz = ureg(Bz)
    if isinstance(Bz, float):
        Bz = Bz * ureg("tesla")
    xs = positions[:, 0]
    ys = positions[:, 1]
    dx = np.ptp(xs)
    dy = np.ptp(ys)
    xs = xs - (xs.min() + dx / 2)
    ys = ys - (ys.min() + dy / 2)
    Ax = -Bz * ys / 2
    Ay = Bz * xs / 2
    A = np.stack([Ax, Ay, np.zeros_like(Ax)], axis=1)
    return A.to("tesla * meter")



import logging
import multiprocessing as mp
from functools import partial
from typing import List, Tuple

import joblib
import numpy as np
import scipy.sparse as sp
from scipy.spatial import ConvexHull, Delaunay, QhullError
from shapely.geometry import MultiLineString
from shapely.ops import orient, polygonize

logger = logging.getLogger("tdgl.finite_volume")


def get_edges(elements: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Finds the edges from a list of triangle indices.

    Args:
        elements: The triangle indices, shape ``(n, 3)``.

    Returns:
        A tuple containing an integer array of edges and a boolean array
        indicating whether each edge on in the boundary.
    """
    edges = np.concatenate([elements[:, e] for e in [(0, 1), (1, 2), (2, 0)]])
    edges = np.sort(edges, axis=1)
    edges, counts = np.unique(edges, return_counts=True, axis=0)
    return edges, counts == 1


def get_edge_lengths(points: np.ndarray, elements: np.ndarray) -> np.ndarray:
    """Returns the lengths of all edges in a triangulation.

    Args:
        points: Vertex coordinates.
        elements: Triangle indices.

    Returns:
        An array of edge lengths.
    """
    edges, _ = get_edges(elements)
    return np.linalg.norm(np.diff(points[edges], axis=1), axis=2).squeeze()


def get_dual_edge_lengths(
    edge_centers: np.ndarray,
    elements: np.ndarray,
    dual_sites: np.ndarray,
    edges: np.ndarray,
) -> np.ndarray:
    """
    Compute the lengths of the dual edges.

    Args:
        edge_centers: The (x, y) coordinates of the edge centers.
        elements: The triangular elements in the tesselation.
        dual_sites: The (x, y) coordinates for the dual mesh (Voronoi sites).
        edges: The edges connecting the sites.

    Returns:
        An array of dual edge lengths.
    """
    # Create a dict with keys corresponding to the edges and values
    # corresponding to the triangles
    edge_to_element = {}
    # Iterate over all elements to create the edge_to_element dict
    edge_element_indices = [[0, 1], [1, 2], [2, 0]]
    for i, element in enumerate(elements):
        for idx in edge_element_indices:
            # Make the array hashable by converting it to a tuple
            edge = tuple(np.sort(element[idx]))
            if edge in edge_to_element:
                edge_to_element[edge].append(i)
            else:
                edge_to_element[edge] = [i]
    dual_lengths = np.zeros(len(edge_centers), dtype=float)
    for i, edge in enumerate(edges):
        indices = edge_to_element[tuple(edge)]
        if len(indices) == 1:  # Boundary edges
            dual_lengths[i] = np.linalg.norm(dual_sites[indices[0]] - edge_centers[i])
        else:  # Inner edges
            dual_lengths[i] = np.linalg.norm(
                dual_sites[indices[0]] - dual_sites[indices[1]]
            )
    return dual_lengths


def generate_voronoi_vertices(
    sites: np.ndarray, elements: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:
    """Compute the vertices of the Voronoi lattice by computing the
    circumcenters of the triangles in the tesselation.

    Args:
        sites: The (x, y) coordinates of the tesselation.
        elements: The triangular elements in the tesselation.

    Returns:
        The x and y coordinates of the Voronoi vertices, as arrays.
    """
    # https://en.wikipedia.org/wiki/Circumscribed_circle#Cartesian_coordinates_2
    # Get the triangle ABC
    # Convert to the coordinate system where A is in the origin
    A = sites[elements[:, 0]]
    B = sites[elements[:, 1]] - A
    C = sites[elements[:, 2]] - A
    # Compute the circumcenter
    D = 2 * B[:, 0] * C[:, 1] - 2 * B[:, 1] * C[:, 0]
    Ux = (C[:, 1] * (B**2).sum(axis=1) - B[:, 1] * (C**2).sum(axis=1)) / D
    Uy = (B[:, 0] * (C**2).sum(axis=1) - C[:, 0] * (B**2).sum(axis=1)) / D
    # Convert back to the initial coordinate system
    return np.array([Ux, Uy]).T + A


def _get_polygon_indices(
    elements: np.ndarray, site_index: int
) -> Tuple[int, np.ndarray]:
    """Helper function for get_voronoi_polygon_indices()."""
    return np.where((elements == site_index).any(axis=1))[0]


def get_voronoi_polygon_indices(
    elements: np.ndarray,
    num_sites: int,
    parallel: bool = True,
    min_sites_for_multiprocessing: int = 10_000,
) -> List[np.ndarray]:
    """Find the polygons surrounding each site.

    Args:
        elements: The triangular elements in the tesselation.
        num_sites: The number of sites
        parallel: If True and the number of sites is greater than
            ``min_sites_for_multiprocessing``, then use multiprocessing.
        min_sites_for_multiprocessing: If ``parallel`` is True and ``num_sites``
            is greater than this value, then use multiprocessing.

    Returns:
        A list of arrays of Voronoi polygon indices.
    """
    # Iterate over all sites and find the triangles that the site belongs to
    # The indices for the triangles are the same as the indices for the
    # Voronoi lattice
    # This is by far the costliest step in Mesh.from_triangulation(),
    # so by default we will use multiprocessing if there are many sites.
    if (
        parallel
        and (ncpus := joblib.cpu_count(only_physical_cores=True)) > 1
        and num_sites > min_sites_for_multiprocessing
    ):
        with mp.Pool(processes=ncpus) as pool:
            results = pool.map(
                partial(_get_polygon_indices, elements), range(num_sites)
            )
        return results
    return [np.where((elements == i).any(axis=1))[0] for i in range(num_sites)]


def compute_voronoi_polygon_areas(
    sites: np.ndarray,
    dual_sites: np.ndarray,
    boundary: np.ndarray,
    edges: np.ndarray,
    boundary_edge_indices: np.ndarray,
    polygons: List[np.ndarray],
) -> Tuple[np.ndarray, List[np.ndarray]]:
    """Compute the areas of the surrounding polygons.

    Areas of boundary points are handled by adding additional points
    on the boundary to make a convex polygon.

    Args:
        sites: The (x, y) coordinates for the sites.
        dual_sites: The (x, y) coordinates of the dual (Voronoi) sites.
        boundary: An array containing all boundary points.
        edges: The edges of the triangles.
        boundary_edge_indices: The edge indices corresponding to the boundary.
        polygons: The polygons in Voronoi diagram.

    Returns:
        An array of areas for each site in the lattice, and a list of
        counterclockwise-oriented Voronoi polygon vertices.
    """

    boundary_set = set(boundary)
    boundary_edges = edges[boundary_edge_indices]
    areas = np.zeros(len(polygons), dtype=float)
    voronoi_sites = []
    for site, polygon in enumerate(polygons):
        # Get the polygon points
        poly = dual_sites[polygon]
        # Polygon vertices may end up very close (e.g. delta = 2e-17) due to floating
        # point errors, so we need to remove near-duplicate vertices.
        _, unique = np.unique(poly.round(decimals=13), axis=0, return_index=True)
        poly = poly[unique]
        if site not in boundary_set:
            areas[site], is_convex = get_convex_polygon_area(poly)
            assert is_convex  # All interior Voronoi cells are convex.
            voronoi_sites.append(orient_convex_polygon(poly))
            continue
        # For points on the boundary, add vertices at the mesh site and the midpoints
        # of the two edges adjacent to the mesh site to complete the Voronoi polygon.
        connected_boundary_edges = boundary_edges[(boundary_edges == site).any(axis=1)]
        # The midpoints of the two edges adjacent to the site
        midpoints = sites[connected_boundary_edges].mean(axis=1)
        # Orient the convex hull of the polygon in a counterclockwise fashion
        coords = orient_convex_polygon(np.concatenate([poly, midpoints], axis=0))
        coords = [tuple(xy) for xy in coords]
        # Insert the central mesh site between the two boundary edge midpoints
        # to ensure the correct ordering of coordinates.
        indices = sorted([coords.index(tuple(mid)) for mid in midpoints])
        if indices[1] == indices[0] + 1:
            # The two boundary edge midpoints are adjacent in the list of coordinates,
            # so insert the mesh site between them.
            coords.insert(indices[1], sites[site])
        else:
            # The boundary edge midpoints are the first and last elements in the
            # list of coordinates, so append the central mesh site to the end.
            if indices[0] != 0:
                # TODO: Decide whether this should be an exception.
                logger.warning(
                    f"Malformed Voronoi cell surrounding boundary site {site}."
                    " Try changing the number of boundary mesh sites using"
                    " Polygon.resample() or Polygon.buffer(eps) where eps"
                    " is 0 or a small positive float."
                )
            coords.append(sites[site])
        poly = np.array(coords)
        areas[site], is_convex = get_convex_polygon_area(poly)
        if not is_convex:
            # If the polygon is non-convex we need to subtract the area of the
            # concave part, which is the triangle formed by the mesh site and
            # the two adjacent boundary edge midpoints.
            triangle_area, is_convex = get_convex_polygon_area(
                np.concatenate([midpoints, [sites[site]]], axis=0)
            )
            assert is_convex  # This is just a triangle, so it must be convex.
            areas[site] -= triangle_area
        voronoi_sites.append(poly)
    return areas, voronoi_sites


def get_convex_polygon_area(coords: np.ndarray) -> Tuple[float, bool]:
    """Compute the area of a convex polygon or the area of its convex hull.

    Note: The vertices do not need to be stored in any specific order.

    Args:
        coords: The (x, y) coordinates of the vertices.

    Returns:
        The area of the polygon or the convex hull, and a bool indicating
        whether the polygon is convex.
    """
    try:
        hull = ConvexHull(coords)
    except QhullError:
        # Handle error when all points lie on a line
        return 0, True
    else:
        is_convex = len(hull.vertices) == len(coords)
        return hull.volume, is_convex


def triangle_areas(points: np.ndarray, triangles: np.ndarray) -> np.ndarray:
    """Calculates the area of each triangle.

    Args:
        points: Shape (n, 2) array of x, y coordinates of vertices
        triangles: Shape (m, 3) array of triangle indices

    Returns:
        Shape (m, ) array of triangle areas
    """
    xy = points[triangles]
    # s1 = xy[:, 2, :] - xy[:, 1, :]
    # s2 = xy[:, 0, :] - xy[:, 2, :]
    # s3 = xy[:, 1, :] - xy[:, 0, :]
    # which can be simplified to
    # s = xy[:, [2, 0, 1]] - xy[:, [1, 2, 0]]  # 3D
    s = xy[:, [2, 0]] - xy[:, [1, 2]]  # 2D
    a = np.linalg.det(s)
    return a * 0.5


def orient_convex_polygon(vertices: np.ndarray) -> np.ndarray:
    """Returns counterclockwise-oriented vertices for a convex polygon.

    Args:
        vertices: The vertex positions (x, y), shape ``(n, 2)``.

    Returns:
        The ``vertices`` sorted counterclockwise.
    """
    # Sort the vertices by the angle between each vertex and some point in the
    # interior of the polygon. Here we use the mean of the vertex positions.
    diffs = vertices - vertices.mean(axis=0)
    return vertices[np.argsort(np.arctan2(diffs[:, 1], diffs[:, 0]))]


def convex_polygon_centroid(points: np.ndarray) -> Tuple[float, float]:
    """Calculates the ``(x, y)`` position of the centroid of a convex polygon.

    Args:
        points: An array of vertex coordinates.

    Returns:
        The ``(x, y)`` position of the centroid of the polygon defined by ``points``.
    """
    # Find a Delaunay triangulation of the polygon
    triangles = Delaunay(points).simplices
    # Find the area and centroid of each triangle
    areas = triangle_areas(points, triangles)
    centroids = points[triangles].mean(axis=1)
    # Return the weighted average of the triangle centroids.
    return np.average(centroids, weights=areas, axis=0)


def get_oriented_boundary(
    points: np.ndarray, boundary_edges: np.ndarray
) -> List[np.ndarray]:
    """Returns arrays of boundary vertex indices, ordered counterclockwise.

    Args:
        points: Shape ``(n, 2)``, float array of vertex coordinates.
        boundary_edges: Shape ``(m, 2)`` integer array of boundary edges.

    Returns:
        A list of arrays of boundary vertex indices (ordered counterclockwise).
        The length of the list will be 1 plus the number of holes in the polygon,
        as each hole has a boundary.
    """
    points_list = [tuple(xy) for xy in points]
    edges = MultiLineString([points[edge, :] for edge in boundary_edges])
    polygons = list(polygonize(edges))
    polygon_indices = []
    for p in polygons:
        polygon = orient(p)
        indices = np.array([points_list.index(xy) for xy in polygon.exterior.coords])
        polygon_indices.append(indices[:-1])
    return polygon_indices


def get_supercurrent(
    psi: np.ndarray, gradient: sp.csr_matrix, edges: np.ndarray
) -> np.ndarray:
    """Compute the supercurrent on the edges.

    Args:
        psi: The value of the complex order parameter.
        gradient: The covariant derivative matrix.
        edges: The indices for the edges.

    Returns:
        The supercurrent at each edge.
    """
    return (psi.conjugate()[edges[:, 0]] * (gradient @ psi)).imag



import numpy as np
import pytest

import tdgl
from tdgl import em
from tdgl.geometry import circle

ureg = tdgl.ureg


@pytest.mark.parametrize("length_units", ["um", "nm"])
@pytest.mark.parametrize("current_units", ["uA", "mA"])
@pytest.mark.parametrize("z_eval", [1, 2, 5, 10])
def test_current_loop(z_eval, current_units, length_units):
    r_loop = 10
    z_loop = 0
    r_eval = 5
    eval_points = 3000
    current = 1

    pl = circle(r_eval, points=1001)
    pl_edge_centers = ((pl + np.roll(pl, -1, axis=0)) / 2)[:-1]
    pl_edges = np.diff(pl, axis=0) * ureg(length_units)
    pl_mesh = tdgl.Polygon(points=pl).make_mesh(min_points=eval_points)
    pl_points = np.concatenate(
        [pl_mesh.sites, z_eval * np.ones((len(pl_mesh.sites), 1))], axis=1
    )
    pl_areas = pl_mesh.areas * ureg(length_units) ** 2

    B = em.current_loop_field(
        pl_points,
        current=current,
        loop_radius=r_loop,
        loop_center=np.array([0, 0, z_loop]),
        num_segments=201,
        length_units=length_units,
        current_units=current_units,
    )

    A = em.current_loop_vector_potential(
        np.concatenate(
            [pl_edge_centers, z_eval * np.ones_like(pl_edge_centers[:, :1])], axis=1
        ),
        current=current,
        loop_radius=r_loop,
        loop_center=np.array([0, 0, z_loop]),
        length_units=length_units,
        current_units=current_units,
    )

    flux_from_B = np.sum(B[:, 2] * pl_areas).to("Phi_0").magnitude
    flux_from_A = np.trapz((A[:, :2] * pl_edges).sum(axis=1)).to("Phi_0").magnitude

    assert np.isclose(flux_from_A, flux_from_B, rtol=1e-2)



from typing import Dict, List, Optional, Sequence, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from scipy import interpolate

from ..visualization import auto_grid, auto_range_iqr
from .solution import Solution


def setup_color_limits(
    dict_of_arrays: Dict[str, np.ndarray],
    vmin: Union[float, None] = None,
    vmax: Union[float, None] = None,
    share_color_scale: bool = False,
    symmetric_color_scale: bool = False,
    auto_range_cutoff: Optional[Union[float, Tuple[float, float]]] = None,
) -> Dict[str, Tuple[float, float]]:
    """Set up color limits (vmin, vmax) for a dictionary of numpy arrays.

    Args:
        dict_of_arrays: Dict of ``{name: array}`` for which to compute color limits.
        vmin: If provided, this vmin will be used for all arrays. If vmin is not None,
            then vmax must also not be None.
        vmax: If provided, this vmax will be used for all arrays. If vmax is not None,
            then vmin must also not be None.
        share_color_scale: Whether to force all arrays to share the same color scale.
            This option is ignored if vmin and vmax are provided.
        symmetric_color_scale: Whether to use a symmetric color scale (vmin = -vmax).
            This option is ignored if vmin and vmax are provided.
        auto_range_cutoff: Cutoff percentile for :func:`tdgl.solution.plot_solution.auto_range_iqr`.

    Returns:
        A dict of ``{name: (vmin, vmax)}``
    """
    if (vmin is not None and vmax is None) or (vmax is not None and vmin is None):
        raise ValueError("If either vmin or max is provided, both must be provided.")
    if vmin is not None:
        return {name: (vmin, vmax) for name in dict_of_arrays}

    if auto_range_cutoff is None:
        clims = {
            name: (np.nanmin(array), np.nanmax(array))
            for name, array in dict_of_arrays.items()
        }
    else:
        clims = {
            name: auto_range_iqr(array, cutoff_percentile=auto_range_cutoff)
            for name, array in dict_of_arrays.items()
        }

    if share_color_scale:
        # All subplots share the same color scale
        global_vmin = np.inf
        global_vmax = -np.inf
        for vmin, vmax in clims.values():
            global_vmin = min(vmin, global_vmin)
            global_vmax = max(vmax, global_vmax)
        clims = {name: (global_vmin, global_vmax) for name in dict_of_arrays}

    if symmetric_color_scale:
        # Set vmin = -vmax
        new_clims = {}
        for name, (vmin, vmax) in clims.items():
            new_vmax = max(vmax, -vmin)
            new_clims[name] = (-new_vmax, new_vmax)
        clims = new_clims

    return clims


def cross_section(
    dataset_coords: np.ndarray,
    dataset_values: np.ndarray,
    cross_section_coords: Union[np.ndarray, Sequence[np.ndarray]],
    interp_method: str = "linear",
) -> Tuple[List[np.ndarray], List[np.ndarray], List[np.ndarray]]:
    """Takes a cross-section of the specified dataset values along
    a path given by the given dataset coordinates.

    Args:
        dataset_coords: A shape (n, 2) array of (x, y) coordinates for the dataset.
        dataset_values: A shape (n, ) array of dataset values of which
            to take a cross-section.
        cross_section_coords: A shape (m, 2) array of (x, y) coordinates specifying
            the cross-section path (or a list of such arrays for multiple
            cross sections).
        interp_method: The interpolation method to use: "nearest", "linear", "cubic".

    Returns:
        A list of coordinate arrays, a list of curvilinear coordinate (path) arrays,
        and a list of cross section values.
    """
    valid_methods = ("nearest", "linear", "cubic")
    if interp_method not in valid_methods:
        raise ValueError(
            f"Interpolation method must be one of {valid_methods} "
            f"(got {interp_method})."
        )
    if interp_method == "nearest":
        interpolator = interpolate.NearestNDInterpolator
    elif interp_method == "linear":
        interpolator = interpolate.LinearNDInterpolator
    else:  # "cubic"
        interpolator = interpolate.CloughTocher2DInterpolator

    if not (isinstance(cross_section_coords, Sequence)):
        cross_section_coords = [cross_section_coords]
    cross_section_coords = [np.asarray(c) for c in cross_section_coords]
    for i, arr in enumerate(cross_section_coords):
        if arr.ndim != 2 or arr.shape[1] != 2:
            raise ValueError(
                f"Invalid shape for coordinate array {i}: {arr.shape}. "
                f"Coordinate arrays must have shape (n, 2)."
            )
    # Calculcate curvilinear cross section coordinates
    paths = []
    for c in cross_section_coords:
        path = np.cumsum(np.sqrt(np.sum(np.diff(c, axis=0) ** 2, axis=1)))
        paths.append(np.concatenate([[0], path], axis=0))
    # Calculate cross sections.
    cross_sections = []
    mask = np.isfinite(dataset_values)
    z_interp = interpolator(dataset_coords[mask], dataset_values[mask])
    for c in cross_section_coords:
        cross_sections.append(z_interp(c[:, 0], c[:, 1]))

    return cross_section_coords, paths, cross_sections


def plot_currents(
    solution: Solution,
    ax: Union[plt.Axes, None] = None,
    dataset: Union[str, None] = None,
    units: Union[str, None] = None,
    cmap: str = "inferno",
    colorbar: bool = True,
    auto_range_cutoff: Optional[Union[float, Tuple[float, float]]] = None,
    symmetric_color_scale: bool = False,
    vmin: Union[float, None] = None,
    vmax: Union[float, None] = None,
    streamplot: bool = True,
    min_stream_amp: float = 0.025,
    cross_section_coords: Union[np.ndarray, Sequence[np.ndarray], None] = None,
    **kwargs,
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plots the sheet current density for a given :class:`tdgl.Solution`.

    Additional keyword arguments are passed to ``plt.subplots()``.

    .. seealso:

        :meth:`tdgl.Solution.plot_currents`

    Args:
        solution: The Solution from which to extract sheet current.
        dataset: The dataset to plot, either ``"supercurrent"`` or
            ``"normal_current"``. ``None`` indicates the total current density.
        ax: Matplotlib axes on which to plot.
        units: Units in which to plot the current density. Defaults to
            ``solution.current_units / solution.device.length_units``.
        cmap: Name of the matplotlib colormap to use.
        colorbar: Whether to add a colorbar to each subplot.
        auto_range_cutoff: Cutoff percentile for :func:`tdgl.solution.plot_solution.auto_range_iqr`.
        symmetric_color_scale: Whether to use a symmetric color scale (vmin = -vmax).
        vmin: Color scale minimum to use for all layers
        vmax: Color scale maximum to use for all layers
        streamplot: Whether to overlay current streamlines on the plot.
        min_stream_amp: Streamlines will not be drawn anywhere the
            current density is less than min_stream_amp * max(current_density).
            This avoids streamlines being drawn where there is no current flowing.
        cross_section_coords: Shape (m, 2) array of (x, y) coordinates for a
            cross-section (or a list of such arrays).

    Returns:
        matplotlib figure and axes
    """
    device = solution.device
    length_units = device.ureg(device.length_units).units
    old_units = device.ureg(f"{solution.current_units} / {device.length_units}").units
    units = units or old_units
    if isinstance(units, str):
        units = device.ureg(units).units
    if dataset is None:
        J = solution.current_density
    elif dataset in ["supercurrent"]:
        J = solution.supercurrent_density
    elif dataset in ["normal_current"]:
        J = solution.normal_current_density
    else:
        raise ValueError(f"Unexpected dataset: {dataset}.")
    if ax is None:
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = ax.get_figure()

    J = J.to(units).magnitude
    Jx = J[:, 0]
    Jy = J[:, 1]
    Jnorm = np.sqrt(Jx**2 + Jy**2)
    x = solution.device.points[:, 0]
    y = solution.device.points[:, 1]
    t = solution.device.triangles
    clabel = "$|\\,\\vec{K}\\,|$" + f" [${units:~L}$]"
    clim = setup_color_limits(
        {"J": Jnorm},
        vmin=vmin,
        vmax=vmax,
        symmetric_color_scale=symmetric_color_scale,
        auto_range_cutoff=auto_range_cutoff,
    )["J"]
    vmin, vmax = clim
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    im = ax.tripcolor(x, y, t, Jnorm, shading="gouraud", cmap=cmap, norm=norm)
    ax.set_title(f"{clabel.split('[')[0].strip()}")
    ax.set_aspect("equal")
    ax.set_xlabel(f"$x$ [${length_units:~L}$]")
    ax.set_ylabel(f"$y$ [${length_units:~L}$]")
    ax.set_xlim(x.min(), x.max())
    ax.set_ylim(y.min(), y.max())
    if cross_section_coords is not None:
        ax_divider = make_axes_locatable(ax)
        cax = ax_divider.append_axes("bottom", size="40%", pad="30%")
        coords, paths, cross_sections = cross_section(
            np.array([x, y]).T,
            Jnorm,
            cross_section_coords,
        )
        for i, (coord, path, cross) in enumerate(zip(coords, paths, cross_sections)):
            color = f"C{i % 10}"
            cross[~device.contains_points(coord)] = np.nan
            ax.plot(*coord.T, "--", color=color, lw=2)
            ax.plot(*coord[0], "o", color=color)
            ax.plot(*coord[-1], "s", color=color)
            cax.plot(path, cross, color=color, lw=2)
            cax.plot(path[0], cross[0], "o", color=color)
            cax.plot(path[-1], cross[-1], "s", color=color)
        cax.grid(True)
        cax.set_xlabel(f"Distance along cut [${length_units:~L}$]")
        cax.set_ylabel(clabel)
    if streamplot:
        xgrid, ygrid, Jgrid = solution.grid_current_density(
            dataset=dataset,
            grid_shape=200,
            method="cubic",
            units=str(units),
            with_units=False,
        )
        Jx, Jy = Jgrid
        J = np.sqrt(Jx**2 + Jy**2)
        xy = np.array([xgrid.ravel(), ygrid.ravel()]).T
        ix = np.where(~solution.device.contains_points(xy)[0])
        ix = np.unravel_index(ix, J.shape)
        Jx[ix] = np.nan
        Jy[ix] = np.nan
        if min_stream_amp is not None:
            cutoff = np.nanmax(J) * min_stream_amp
            Jx[J < cutoff] = np.nan
            Jy[J < cutoff] = np.nan
        ax.streamplot(xgrid, ygrid, Jx, Jy, color="w", density=1, linewidth=0.75)
    if colorbar:
        cbar = fig.colorbar(im, ax=ax, orientation="vertical")
        cbar.set_label(clabel)
    return fig, ax


def plot_field_at_positions(
    solution: Solution,
    positions: np.ndarray,
    zs: Optional[Union[float, np.ndarray]] = None,
    vector: bool = False,
    units: Union[str, None] = None,
    grid_shape: Union[int, Tuple[int, int]] = (200, 200),
    grid_method: str = "cubic",
    cmap: str = "cividis",
    colorbar: bool = True,
    auto_range_cutoff: Optional[Union[float, Tuple[float, float]]] = None,
    share_color_scale: bool = False,
    symmetric_color_scale: bool = False,
    vmin: Union[float, None] = None,
    vmax: Union[float, None] = None,
    cross_section_coords: Optional[Union[float, List[float]]] = None,
    **kwargs,
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plots the Biot-Savart field (either all three components or just the
    z component) at a given set of positions (x, y, z) outside of the device.

    .. note::

        This function plots only the field due to currents flowing in the device.
        It does not include the applied field.

    .. seealso:

        :meth:`tdgl.Solution.plot_field_at_positions`

    Additional keyword arguments are passed to ``plt.subplots()``. This function first
    evaluates the field at ``positions``, then interpolates the resulting fields to a
    rectangular grid for plotting.

    Args:
        solution: The Solution from which to extract fields.
        positions: Shape (m, 2) array of (x, y) coordinates, or (m, 3) array of (x, y, z)
            coordinates at which to calculate the magnetic field.
        zs: z coordinates at which to calculate the field. If positions has shape (m, 3), then
            this argument is not allowed. If zs is a scalar, then the fields are calculated in
            a plane parallel to the x-y plane. If zs is an array, then it must be same length
            as positions.
        vector: Whether to plot the full vector magnetic field or just the z component.
        units: Units in which to plot the fields. Defaults to ``solution.field_units``.
        grid_shape: Shape of the desired rectangular grid. If a single integer ``n``
            is given, then the grid will be square, shape ``(n, n)``.
        grid_method: Interpolation method to use (see :func:`scipy.interpolate.griddata`).
        max_cols: Maximum number of columns in the grid of subplots.
        cmap: Name of the matplotlib colormap to use.
        colorbar: Whether to add a colorbar to each subplot.
        auto_range_cutoff: Cutoff percentile for :func:`tdgl.solution.plot_solution.auto_range_iqr`.
        share_color_scale: Whether to force all layers to use the same color scale.
        symmetric_color_scale: Whether to use a symmetric color scale (vmin = -vmax).
        vmin: Color scale minimum to use for all layers
        vmax: Color scale maximum to use for all layers
        cross_section_coords: Shape (m, 2) array of (x, y) coordinates for a
            cross-section (or a list of such arrays).

    Returns:
        matplotlib figure and axes
    """
    device = solution.device
    # Length units from the Device
    length_units = device.ureg(device.length_units).units
    # The units the fields are currently in
    old_units = device.ureg(solution.field_units).units
    # The units we want to convert to
    if units is None:
        units = old_units
    if isinstance(units, str):
        units = device.ureg(units).units
    fields = solution.field_at_position(
        positions,
        zs=zs,
        vector=vector,
        units=units,
        with_units=False,
    )
    if fields.ndim == 1:
        fields = fields[:, np.newaxis]
    if vector:
        num_subplots = 3
    else:
        num_subplots = 1
    fig, axes = auto_grid(num_subplots, **kwargs)
    if not isinstance(axes, (list, np.ndarray)):
        axes = [axes]
    x, y, *_ = positions.T
    xs = np.linspace(x.min(), x.max(), grid_shape[1])
    ys = np.linspace(y.min(), y.max(), grid_shape[0])
    xgrid, ygrid = np.meshgrid(xs, ys)
    # Shape grid_shape or (grid_shape + (3, ))
    fields = interpolate.griddata(
        positions[:, :2],
        fields,
        (xgrid, ygrid),
        method=grid_method,
    )
    clabels = [f"{label} [${units:~L}$]" for label in ["$H_x$ ", "$H_y$ ", "$H_z$ "]]
    if "[mass]" in units.dimensionality:
        # We want flux density, B = mu0 * H
        clabels = ["$\\mu_0$" + clabel for clabel in clabels]
    if not vector:
        clabels = clabels[-1:]
    fields_dict = {label: fields[:, :, i] for i, label in enumerate(clabels)}
    clim_dict = setup_color_limits(
        fields_dict,
        vmin=vmin,
        vmax=vmax,
        share_color_scale=share_color_scale,
        symmetric_color_scale=symmetric_color_scale,
        auto_range_cutoff=auto_range_cutoff,
    )
    for ax, label in zip(fig.axes, clabels):
        field = fields_dict[label]
        layer_vmin, layer_vmax = clim_dict[label]
        norm = mpl.colors.Normalize(vmin=layer_vmin, vmax=layer_vmax)
        im = ax.pcolormesh(xgrid, ygrid, field, shading="auto", cmap=cmap, norm=norm)
        ax.set_title(f"{label.split('[')[0].strip()}")
        ax.set_aspect("equal")
        ax.set_xlabel(f"$x$ [${length_units:~L}$]")
        ax.set_ylabel(f"$y$ [${length_units:~L}$]")
        ax.set_xlim(xgrid.min(), xgrid.max())
        ax.set_ylim(ygrid.min(), ygrid.max())
        if cross_section_coords is not None:
            ax_divider = make_axes_locatable(ax)
            cax = ax_divider.append_axes("bottom", size="40%", pad="30%")
            coords, paths, cross_sections = cross_section(
                np.array([xgrid.ravel(), ygrid.ravel()]).T,
                field.ravel(),
                cross_section_coords=cross_section_coords,
            )
            for i, (coord, path, cross) in enumerate(
                zip(coords, paths, cross_sections)
            ):
                color = f"C{i % 10}"
                ax.plot(*coord.T, "--", color=color, lw=2)
                ax.plot(*coord[0], "o", color=color)
                ax.plot(*coord[-1], "s", color=color)
                cax.plot(path, cross, color=color, lw=2)
                cax.plot(path[0], cross[0], "o", color=color)
                cax.plot(path[-1], cross[-1], "s", color=color)
            cax.grid(True)
            cax.set_xlabel(f"Distance along cut [${length_units:~L}$]")
            cax.set_ylabel(label)
        if colorbar:
            cbar = fig.colorbar(im, ax=ax, orientation="vertical")
            cbar.set_label(label)
    return fig, axes


def plot_order_parameter(
    solution: Solution,
    squared: bool = False,
    mag_cmap: str = "viridis",
    phase_cmap: str = "twilight_shifted",
    shading: str = "gouraud",
    **kwargs,
) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
    """Plots the magnitude (or the magnitude squared) and
    phase of the complex order parameter, :math:`\\psi=|\\psi|e^{i\\theta}`.

    .. seealso:

        :meth:`tdgl.Solution.plot_order_parameter`

    Args:
        solution: The solution for which to plot the order parameter.
        squared: Whether to plot the magnitude squared, :math:`|\\psi|^2`.
        mag_cmap: Name of the colormap to use for the magnitude.
        phase_cmap: Name of the colormap to use for the phase.
        shading: May be ``"flat"`` or ``"gouraud"``. The latter does some interpolation.

    Returns:
        matplotlib Figure and an array of two Axes objects.
    """
    kwargs.setdefault("figsize", (8, 3))
    kwargs.setdefault("constrained_layout", True)
    device = solution.device
    psi = solution.tdgl_data.psi
    mag = np.abs(psi)
    psi_label = "$|\\psi|$"
    if squared:
        mag = mag**2
        psi_label = "$|\\psi|^2$"
    phase = np.angle(psi) / np.pi
    points = device.points
    triangles = device.triangles
    fig, axes = plt.subplots(1, 2, **kwargs)
    im = axes[0].tripcolor(
        points[:, 0],
        points[:, 1],
        mag,
        triangles=triangles,
        vmin=0,
        vmax=1,
        cmap=mag_cmap,
        shading=shading,
    )
    cbar = fig.colorbar(im, ax=axes[0])
    cbar.set_label(psi_label)
    im = axes[1].tripcolor(
        points[:, 0],
        points[:, 1],
        phase,
        triangles=triangles,
        vmin=-1,
        vmax=1,
        cmap=phase_cmap,
        shading=shading,
    )
    cbar = fig.colorbar(im, ax=axes[1])
    cbar.set_label("$\\theta / \\pi$")
    length_units = device.ureg(device.length_units).units
    for ax in axes:
        ax.set_aspect("equal")
        ax.set_xlabel(f"$x$ [${length_units:~L}$]")
        ax.set_ylabel(f"$y$ [${length_units:~L}$]")
    return fig, axes


def plot_vorticity(
    solution: Solution,
    ax: Union[plt.Axes, None] = None,
    cmap: str = "coolwarm",
    units: Union[str, None] = None,
    auto_range_cutoff: Optional[Union[float, Tuple[float, float]]] = None,
    symmetric_color_scale: bool = True,
    vmin: Union[float, None] = None,
    vmax: Union[float, None] = None,
    shading: str = "gouraud",
    **kwargs,
):
    """Plots the vorticity in the film:
    :math:`\\mathbf{\\omega}=\\mathbf{\\nabla}\\times\\mathbf{K}`.

    .. seealso:

        :meth:`tdgl.Solution.plot_vorticity`

    Args:
        solution: The solution for which to plot the vorticity.
        ax: Matplotlib axes on which to plot.
        cmap: Name of the matplotlib colormap to use.
        units: The units in which to plot the vorticity. Must have dimensions of
            [current] / [length]^2.
        auto_range_cutoff: Cutoff percentile for :func:`tdgl.solution.plot_solution.auto_range_iqr`.
        symmetric_color_scale: Whether to use a symmetric color scale (vmin = -vmax).
        vmin: Color scale minimum.
        vmax: Color scale maximum.
        shading: May be ``"flat"`` or ``"gouraud"``. The latter does some interpolation.

    Returns:
        matplotlib Figure and and Axes.
    """
    if ax is None:
        kwargs.setdefault("constrained_layout", True)
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = ax.get_figure()
    ax.set_aspect("equal")
    device = solution.device
    points = device.points
    triangles = device.triangles
    length_units = device.ureg(device.length_units).units
    if units is None:
        units = solution.vorticity.units
    else:
        units = device.ureg(units)
    v = solution.vorticity.to(units).m
    clim = setup_color_limits(
        {"v": v},
        vmin=vmin,
        vmax=vmax,
        symmetric_color_scale=symmetric_color_scale,
        auto_range_cutoff=auto_range_cutoff,
    )["v"]
    vmin, vmax = clim
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    x, y = points[:, 0], points[:, 1]
    im = ax.tripcolor(
        x,
        y,
        v,
        triangles=triangles,
        cmap=cmap,
        norm=norm,
        shading=shading,
    )
    cbar = fig.colorbar(im, ax=ax)
    ax.set_title("$\\vec{\\omega}=\\vec{\\nabla}\\times\\vec{K}$")
    ax.set_aspect("equal")
    ax.set_xlabel(f"$x$ [${length_units:~L}$]")
    ax.set_ylabel(f"$y$ [${length_units:~L}$]")
    cbar.set_label(f"$\\vec{{\\omega}}\\cdot\\hat{{z}}$ [${units:~L}$]")
    ax.set_xlim(x.min(), x.max())
    ax.set_ylim(y.min(), y.max())
    return fig, ax


def plot_scalar_potential(
    solution: Solution,
    ax: Union[plt.Axes, None] = None,
    cmap: str = "magma",
    auto_range_cutoff: Optional[Union[float, Tuple[float, float]]] = None,
    vmin: Union[float, None] = None,
    vmax: Union[float, None] = None,
    shading: str = "gouraud",
    **kwargs,
):
    """Plots the scalar potential :math:`\\mu(\\mathbf{r})` in the film.

    .. seealso:

        :meth:`tdgl.Solution.plot_scalar_potential`

    Args:
        solution: The solution for which to plot the scalar potential.
        ax: Matplotlib axes on which to plot.
        cmap: Name of the matplotlib colormap to use.
        auto_range_cutoff: Cutoff percentile for :func:`tdgl.solution.plot_solution.auto_range_iqr`.
        vmin: Color scale minimum.
        vmax: Color scale maximum.
        shading: May be ``"flat"`` or ``"gouraud"``. The latter does some interpolation.

    Returns:
        matplotlib Figure and and Axes.
    """
    if ax is None:
        kwargs.setdefault("constrained_layout", True)
        fig, ax = plt.subplots(**kwargs)
    else:
        fig = ax.get_figure()
    ax.set_aspect("equal")
    device = solution.device
    points = device.points
    triangles = device.triangles
    length_units = device.ureg(device.length_units).units
    mu = solution.tdgl_data.mu
    mu = mu - np.nanmin(mu)
    clim = setup_color_limits(
        {"mu": mu},
        vmin=vmin,
        vmax=vmax,
        auto_range_cutoff=auto_range_cutoff,
    )["mu"]
    vmin, vmax = clim
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)
    x, y = points[:, 0], points[:, 1]
    im = ax.tripcolor(
        x,
        y,
        mu,
        triangles=triangles,
        cmap=cmap,
        norm=norm,
        shading=shading,
    )
    cbar = fig.colorbar(im, ax=ax)
    ax.set_title("$\\mu/v_0$")
    ax.set_aspect("equal")
    ax.set_xlabel(f"$x$ [${length_units:~L}$]")
    ax.set_ylabel(f"$y$ [${length_units:~L}$]")
    cbar.set_label("$\\mu/v_0$")
    ax.set_xlim(x.min(), x.max())
    ax.set_ylim(y.min(), y.max())
    return fig, ax


def _patch_docstring(func):
    other_func = getattr(Solution, func.__name__)
    other_func.__doc__ = (
        other_func.__doc__
        + "\n\n"
        + "\n".join(
            [line for line in func.__doc__.split("\n    ") if "solution:" not in line]
        )
    )
    annotations = func.__annotations__.copy()
    _ = annotations.pop("solution", None)
    other_func.__annotations__.update(annotations)


for func in (
    plot_currents,
    plot_field_at_positions,
    plot_order_parameter,
    plot_scalar_potential,
    plot_vorticity,
):
    _patch_docstring(func)



import dataclasses
from typing import Any, Dict, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np

from ..finite_volume.mesh import Mesh


def get_data_range(h5file: h5py.File) -> Tuple[int, int]:
    """Returns the minimum and maximum solve steps in the file."""
    keys = np.asarray([int(key) for key in h5file["data"]])
    return np.min(keys), np.max(keys)


def load_state_data(h5file: h5py.File, step: int) -> Dict[str, Any]:
    """Returns a dict of state data for the given solve step."""
    return dict(h5file["data"][str(step)].attrs)


def array_safe_equals(a: Any, b: Any) -> bool:
    """Check if a and b are equal, even if they are numpy arrays."""
    if a is b:
        return True
    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return a.shape == b.shape and np.allclose(a, b)
    try:
        return a == b
    except TypeError:
        return NotImplemented


def dataclass_equals(dc1: Any, dc2: Any) -> bool:
    """Check if two dataclasses that may hold numpy arrays are equal."""
    if dc1 is dc2:
        return True
    if dc1.__class__ is not dc2.__class__:
        return NotImplemented
    t1 = dataclasses.astuple(dc1)
    t2 = dataclasses.astuple(dc2)
    return all(array_safe_equals(a1, a2) for a1, a2 in zip(t1, t2))


def get_edge_quantity_data(
    quantity_on_edges: np.ndarray, mesh: Mesh
) -> Tuple[np.ndarray, np.ndarray, Tuple[float, float]]:
    """Returns the value of a vector quantity living on the edges of the mesh,
    evaluated at sites in the mesh.

    Args:
        quantity_on_edges: An array of quantity values on the mesh edges.
        mesh: The :class:`tdgl.finite_volume.mesh.Mesh` instance.

    Returns:
        The magnitude and directions of the quantity on the sites, and a tuple
        of the (min, max) of the data.
    """
    directions = mesh.get_quantity_on_site(quantity_on_edges)
    norm = np.linalg.norm(directions, axis=1)
    directions /= np.maximum(norm, 1e-12)[:, np.newaxis]
    return norm, directions, (np.min(norm), np.max(norm))


@dataclasses.dataclass(eq=False)
class TDGLData:
    """A container for raw data from the TDGL solver at a single solve step.

    Args:
        step: The solver iteration.
        epsilon: The disorder parameter. :math:`\\epsilon<1` weakens the
            order parameter.
        psi: The complex order parameter at each site in the mesh.
        mu: The scalar potential at each site in the mesh.
        applied_vector_potential: The applied vector potential at each edge in the mesh.
        induced_vector_potential: The induced vector potential at each edge in the mesh.
        supercurrent: The supercurrent density at each edge in the mesh.
        normal_current: The normal density at each edge in the mesh.
        state: The solver state for the current iteration.
    """

    step: int
    epsilon: np.ndarray
    psi: np.ndarray
    mu: np.ndarray
    applied_vector_potential: np.ndarray
    induced_vector_potential: np.ndarray
    supercurrent: np.ndarray
    normal_current: np.ndarray
    state: Dict[str, Any]

    @staticmethod
    def from_hdf5(h5file: Union[h5py.File, h5py.Group], step: int) -> "TDGLData":
        """Load a :class:`TDGLData` instance from an output :class:`h5py.File`.

        Args:
            h5file: An open HDF5 output file.
            step: The solver iteration for which to load data.

        Returns:
            A :class:`TDGLData` instance containing data for the requested solve step.
        """
        step = str(step)

        def get(key, default=None):
            if key in ["step"]:
                return int(step)
            if key in ["state"]:
                return load_state_data(h5file, step)
            if key in h5file:
                return np.asarray(h5file[key])
            if key in h5file["data"][step]:
                return np.array(h5file["data"][step][key])
            return default

        return TDGLData(
            **{field.name: get(field.name) for field in dataclasses.fields(TDGLData)}
        )

    def to_hdf5(self, h5group: h5py.Group) -> None:
        """Save a :class:`TDGLData` instance to an :class:`h5py.Group`.

        Args:
            h5group: An open :class:`h5py.Group` in which to save the data.
        """
        group = h5group.create_group(str(self.step))
        for key, value in dataclasses.asdict(self).items():
            if key in ["step"]:
                continue
            if key in ["state"]:
                group.attrs.update(value)
            else:
                group[key] = value

    def __eq__(self, other: Any) -> bool:
        return dataclass_equals(self, other)


@dataclasses.dataclass(eq=False)
class DynamicsData:
    """A container for the measured dynamics of a TDGL solution,
    measured at each time step in the simulation.

    Args:
        dt: The solver time step, :math:`\\Delta t^{n}`.
        time: The solver time, a derived attribute which is equal to the cumulative
            sum of the time step.
        mu: The electric potential, :math:`\\mu`.
        theta: The phase of the order parameter, :math:`\\theta=\\arg\\psi`
        screening_iterations: The number of screening iterations performed at each
            time step.
    """

    dt: np.ndarray
    time: np.ndarray = dataclasses.field(init=False)
    mu: Union[np.ndarray, None] = None
    theta: Union[np.ndarray, None] = None
    screening_iterations: Union[np.ndarray, None] = None

    def __post_init__(self):
        self.time = np.cumsum(self.dt)

    def time_slice(self, tmin: float = -np.inf, tmax: float = np.inf) -> np.ndarray:
        """Returns the integer indices corresponding to the specified time window.

        Args:
            tmin: The minimum of the time window.
            tmax: The maximum of the time window.

        Returns:
            An array of indices for the time window.
        """
        ts = self.time
        (indices,) = np.where((ts >= tmin) & (ts <= tmax))
        return indices

    def closest_time(self, time: float) -> int:
        """Returns the index of the time step closest to ``time``.

        Args:
            time: The time for which to find the index.

        Returns:
            The index of the time step closest to ``time``.
        """
        return np.argmin(np.abs(self.time - time))

    def voltage(self, i: int = 0, j: int = 1) -> np.ndarray:
        """Returns the voltage, i.e., the electric potential difference
        between probe points ``i`` and ``j``, as a function of time.

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.

        Returns:
            The voltage :math:`V_{ij}(t)=\\mu_i(t)-\\mu_j(t)`
        """
        if self.mu is None:
            raise ValueError("No voltage data available.")
        if self.mu.shape[0] == 1:
            raise ValueError("The solution has only one probe point.")
        return self.mu[i] - self.mu[j]

    def phase_difference(self, i: int = 0, j: int = 1) -> np.ndarray:
        """Returns the phase difference between probe points ``i`` and ``j``
        as a function of time.

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.

        Returns:
            The phase difference :math:`\\Delta\\theta_{ij}(t)=\\theta_i(t)-\\theta_j(t)`,
            where :math:`\\theta=\\arg\\psi`.
        """
        if self.theta is None:
            raise ValueError("No phase data available.")
        if self.theta.shape[0] == 1:
            raise ValueError("The solution has only one probe point.")
        return self.theta[i] - self.theta[j]

    def mean_voltage(
        self, i: int = 0, j: int = 1, tmin: float = -np.inf, tmax: float = np.inf
    ) -> float:
        """Returns the time-averaged voltage :math:`\\langle \\Delta\\mu \\rangle`
        over the specified time interval.

        .. math::

            \\langle V_{i,j} \\rangle =
            \\frac{\\sum_n V_{i,j}^{n}\\cdot\\Delta t^{n}}{\\sum_n\\Delta t^{n}}

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.
            tmin: The minimum of the time window over which to average.
            tmax: The maximum of the time window over which to average.

        Returns:
            The time-averaged voltage over the specified time window.
        """
        if self.mu is None:
            raise ValueError("No voltage data available.")
        indices = self.time_slice(tmin, tmax)
        return np.average(self.voltage(i, j)[indices], weights=self.dt[indices])

    def resample(self, num_points: Union[int, None] = None) -> "DynamicsData":
        """Re-sample the dynamics to a uniform grid using linear interpolation.

        Args:
            num_points: The number of points to interpolate to.

        Returns:
            A new :class:`DynamicsData` instance with the re-sampled data.
        """
        time = self.time
        if num_points is None:
            num_points = len(time)
        ts = np.linspace(time.min(), time.max(), num_points)
        mu = theta = None
        if self.mu is not None:
            mu = np.array([np.interp(ts, time, val) for val in self.mu])
        if self.theta is not None:
            theta = np.array([np.interp(ts, time, val) for val in self.theta])
        return DynamicsData(dt=(ts[1] - ts[0]) * np.ones_like(ts), mu=mu, theta=theta)

    def plot(
        self,
        i: int = 0,
        j: int = 1,
        tmin: float = -np.inf,
        tmax: float = +np.inf,
        grid: bool = True,
        mean_voltage: bool = True,
        labels: bool = True,
        legend: bool = False,
    ) -> Tuple[plt.Figure, Sequence[plt.Axes]]:
        """Plot the voltage and phase difference over the specified time window.

        Args:
            i: Index for the first probe point.
            j: Index for the second probe point.
            tmin: The minimum of the time window to plot.
            tmax: The maximum of the time window to plot.
            grid: Whether to add grid lines to the plots.
            mean_voltage: Whether to plot a horizontal line at the mean voltage.
            labels: Whether to include axis labels.
            legend: Whether to include a legend.

        Returns:
            matplotlib figure and axes.
        """
        fig, axes = plt.subplots(2, 1, sharex=True)
        ax, bx = axes
        ax.grid(grid)
        bx.grid(grid)
        ts = self.time
        vs = self.voltage(i, j)
        phases = np.unwrap(self.phase_difference(i, j)) / np.pi
        indices = self.time_slice(tmin, tmax)
        ax.plot(ts[indices], vs[indices])
        if mean_voltage:
            ax.axhline(
                self.mean_voltage(i=i, j=j, tmin=tmin, tmax=tmax),
                label="Mean voltage",
                color="k",
                ls="--",
            )
        bx.plot(ts[indices], phases[indices])
        if labels:
            ax.set_ylabel(f"Voltage\n$\\Delta\\mu_{{{i},{j}}}$ [$V_0$]")
            bx.set_xlabel("Time, $t$ [$\\tau_0$]")
            bx.set_ylabel(f"Phase difference\n$\\Delta\\theta_{{{i},{j}}}/\\pi$")
        if legend:
            ax.legend(loc=0)
        return fig, axes

    def plot_dt(
        self,
        tmin: float = -np.inf,
        tmax: float = +np.inf,
        grid: bool = True,
        labels: bool = True,
        **histogram_kwargs,
    ) -> Tuple[plt.Figure, Tuple[plt.Axes, plt.Axes]]:
        """Plots the time step :math:`\\Delta t^{n}` vs. time and a histogram of
        :math:`\\Delta t^{n}`.

        Args:
            tmin: The minimum of the time window to plot.
            tmax: The maximum of the time window to plot.
            grid: Whether to add grid lines to the plots.
            labels: Whether to include axis labels.
            histogram_kwargs: Passed to plt.Axes.hist().

        Returns:
            matplotlib figure and two axes.
        """
        fig, (ax, bx) = plt.subplots(1, 2, gridspec_kw=dict(width_ratios=[2, 1]))
        ax.sharey(bx)
        ax.grid(grid)
        bx.grid(grid)
        ts = self.time
        indices = self.time_slice(tmin, tmax)
        ax.plot(ts[indices], self.dt[indices])
        histogram_kwargs = histogram_kwargs.copy()
        histogram_kwargs.setdefault("bins", 101)
        histogram_kwargs.setdefault("density", True)
        histogram_kwargs["orientation"] = "horizontal"
        bx.hist(self.dt[indices], **histogram_kwargs)
        if labels:
            ax.set_xlabel("Time, $t$ [$\\tau_0$]")
            ax.set_ylabel("Time step, $\\Delta t$ [$\\tau_0$]")
            if histogram_kwargs.get("density", False):
                bx.set_xlabel("Density")
            else:
                bx.set_xlabel("Counts per bin")
        fig.tight_layout()
        return fig, (ax, bx)

    @staticmethod
    def from_hdf5(
        h5file: Union[h5py.File, h5py.Group],
        step_min: Union[int, None] = None,
        step_max: Union[int, None] = None,
    ) -> "DynamicsData":
        """Load a :class:`DynamicsData` instance from an output :class:`h5py.File`.

        Args:
            h5file: An open HDF5 output file.
            step_min: The minimum solve step to load.
            step_max: The maximum solve step to load.

        Returns:
            A new :class:`DynamicsData` instance.
        """
        iterations = None
        if "theta" in h5file:
            # Load from DynamicsData.to_hdf5()
            dt = np.array(h5file["dt"])
            mu = screening_iterations = None
            theta = np.array(h5file["theta"])
            if "mu" in h5file:
                mu = np.array(h5file["mu"])
            if "screening_iterations" in h5file:
                iterations = np.array(h5file["screening_iterations"])
        else:
            dts = []
            mus = []
            thetas = []
            screening_iterations = []
            if step_min is None:
                step_min, step_max = get_data_range(h5file)
            for i in range(step_min, step_max + 1):
                grp = h5file[f"data/{i}"]
                if "running_state" not in grp:
                    continue
                grp = grp["running_state"]
                dts.append(np.array(grp["dt"]))
                if "mu" in grp:
                    mus.append(np.array(grp["mu"]))
                if "theta" in grp:
                    thetas.append(np.array(grp["theta"]))
                if "screening_iterations" in grp:
                    screening_iterations.append(np.array(grp["screening_iterations"]))
            dt = np.concatenate(dts)
            mask = dt > 0
            dt = dt[mask]
            mu = theta = iterations = None
            if mus:
                mu = np.concatenate(mus, axis=1)[..., mask]
            if thetas:
                theta = np.concatenate(thetas, axis=1)[..., mask]
            if screening_iterations:
                iterations = np.concatenate(screening_iterations)[mask]
        return DynamicsData(
            dt,
            mu=mu,
            theta=theta,
            screening_iterations=iterations,
        )

    def to_hdf5(self, h5group: h5py.Group) -> None:
        """Save a :class:`DynamicsData` instance to an :class:`h5py.Group`.

        Args:
            h5group: An open :class:`h5py.Group` in which to save the data.
        """
        h5group["dt"] = self.dt
        if self.mu is not None:
            h5group["mu"] = self.mu
        if self.theta is not None:
            h5group["theta"] = self.theta
        if self.screening_iterations is not None:
            h5group["screening_iterations"] = self.screening_iterations

    def __eq__(self, other: Any) -> bool:
        return dataclass_equals(self, other)



import pickle

import numpy as np
import pytest

import tdgl


def test_function_repr():
    def func(x, *args, **kwargs):
        pass

    print(tdgl.parameter.function_repr(func))

    def func(x, a: bool = True, b=None):
        pass

    print(tdgl.parameter.function_repr(func))

    def func(x, *, a: int, b: float, c=None):
        pass

    print(tdgl.parameter.function_repr(func))


def func_2d(x, y, sigma=1):
    return np.exp(-(x**2 + y**2) / (2 * sigma**2))


def func_3d(x, y, z, sigma=1):
    return np.exp(-(x**2 + y**2 + z**2) / (2 * sigma**2))


def test_parameter_against_func():
    def func(x, y, sigma=1):
        return np.exp(-(x**2 + y**2) / (2 * sigma**2))

    param = tdgl.Parameter(func, sigma=2)
    print(param)

    x = np.random.rand(100)
    y = np.random.rand(100)

    assert np.array_equal(func(x, y, sigma=2), param(x, y))

    assert param != tdgl.Parameter(func, sigma=1)

    def func2(x, y, sigma=1):
        return np.exp(-np.abs(x + y) / (2 * sigma**2))

    assert param != tdgl.Parameter(func2, sigma=1)


@pytest.mark.parametrize(
    "func, args",
    [
        (func_2d, (np.random.rand(100), np.random.rand(100))),
        (func_3d, (np.random.rand(100), np.random.rand(100), np.random.rand(100))),
    ],
)
def test_parameter_math(func, args):

    param1 = tdgl.Parameter(func, sigma=10)
    param2 = tdgl.Parameter(func, sigma=0.1)
    print(param1, param2)

    assert np.array_equal((param1**2)(*args), func(*args, sigma=10) ** 2)
    assert np.array_equal((2 * param1)(*args), 2 * func(*args, sigma=10))
    assert np.array_equal((1 + param1)(*args), 1 + func(*args, sigma=10))
    assert np.array_equal((1 - param1)(*args), 1 - func(*args, sigma=10))
    assert np.array_equal((2**param1)(*args), 2 ** func(*args, sigma=10))
    assert np.array_equal((1 / (10 + param1))(*args), 1 / (10 + func(*args, sigma=10)))

    assert np.array_equal(
        (param1 + param2)(*args), func(*args, sigma=10) + func(*args, sigma=0.1)
    )

    assert np.array_equal(
        (param1 - param2)(*args),
        func(*args, sigma=10) - func(*args, sigma=0.1),
    )

    assert np.array_equal(
        (param1 * param2)(*args),
        func(*args, sigma=10) * func(*args, sigma=0.1),
    )

    assert np.array_equal(
        (param1 / param2)(*args),
        func(*args, sigma=10) / func(*args, sigma=0.1),
    )

    assert np.array_equal(
        (param1**param2)(*args),
        func(*args, sigma=10) ** func(*args, sigma=0.1),
    )

    assert param1 == param1
    assert param2 == param2
    assert param1 != param2

    p3 = param1 * param2
    assert p3 == p3

    assert (param1 * param2) == (param1 * param2)
    assert (param1**2) != param1
    assert (param1**2) != param1 * param2
    print(param1**2)
    print(param1 * param2 + param1)
    print(param1 * (param2 + param1))
    print(2 * param2, param1 / 5)


def test_bad_composite_parameter():
    with pytest.raises(TypeError):
        _ = tdgl.parameter.CompositeParameter(1, 2, "+")

    p1 = tdgl.Parameter(func_2d, sigma=10)
    p2 = tdgl.Parameter(func_2d, sigma=5)

    with pytest.raises(ValueError):
        _ = tdgl.parameter.CompositeParameter(p1, p2, "<<")


@pytest.mark.parametrize("func", [func_2d, func_3d])
def test_pickle_parameter(func):

    param1 = tdgl.Parameter(func, sigma=10)
    param2 = tdgl.Parameter(func, sigma=0.1)

    assert pickle.loads(pickle.dumps(param1)) == param1
    assert pickle.loads(pickle.dumps(param1)) != param2

    assert pickle.loads(pickle.dumps(param1 * param2)) == (param1 * param2)
    assert pickle.loads(pickle.dumps(param1 - param2)) == (param1 - param2)
    assert pickle.loads(pickle.dumps(param1 * param2)) != (param1 / param2)


def test_bad_parameters():
    def func(a, x, y, b=0):
        pass

    with pytest.raises(ValueError):
        _ = tdgl.Parameter(func, b=0)

    def func(x, y, a, z):
        pass

    with pytest.raises(ValueError):
        _ = tdgl.Parameter(func)

    def func(x, y, z, a):
        pass

    with pytest.raises(ValueError):
        _ = tdgl.Parameter(func)

    def func(x, y, a=0, b=0):
        pass

    with pytest.raises(ValueError):
        _ = tdgl.Parameter(func, a=0, c=None)



import numpy as np

from ..em import uniform_Bz_vector_potential, ureg
from ..parameter import Parameter


def constant_field_vector_potential(
    x,
    y,
    z,
    *,
    Bz: float,
    field_units: str = "mT",
    length_units: str = "um",
):
    if z.ndim == 0:
        z = z * np.ones_like(x)
    positions = np.array([x.squeeze(), y.squeeze(), z.squeeze()]).T
    positions = (positions * ureg(length_units)).to("m").magnitude
    Bz = Bz * ureg(field_units)
    A = uniform_Bz_vector_potential(positions, Bz)
    return A.to(f"{field_units} * {length_units}").magnitude


def ConstantField(
    value: float = 0, field_units: str = "mT", length_units: str = "um"
) -> Parameter:
    """Returns a Parameter that computes a constant as a function of ``x, y, z``.
    Args:
        value: The constant value of the field.
    Returns:
        A Parameter that returns ``value`` at all ``x, y, z``.
    """
    return Parameter(
        constant_field_vector_potential,
        Bz=float(value),
        field_units=field_units,
        length_units=length_units,
    )



from typing import List, Sequence, Tuple, Union

import h5py
import matplotlib.pyplot as plt
import numpy as np

from ..geometry import close_curve
from .edge_mesh import EdgeMesh
from .util import (
    compute_voronoi_polygon_areas,
    convex_polygon_centroid,
    generate_voronoi_vertices,
    get_edges,
    get_voronoi_polygon_indices,
)


class Mesh:
    """A triangular mesh of a simply- or multiply-connected polygon.

    .. tip::

        Use :meth:`Mesh.from_triangulation` to create a new mesh from a triangulation.

    Args:
        sites: The (x, y) coordinates of the mesh vertices.
        elements: A list of triplets that correspond to the indices of he vertices that
            form a triangle. [[0, 1, 2], [0, 1, 3]] corresponds to a triangle
            connecting vertices 0, 1, and 2 and another triangle connecting vertices
            0, 1, and 3.
        boundary_indices: Indices corresponding to the boundary.
        areas: The areas corresponding to the sites.
        dual_sites: The (x, y) coordinates of the dual (Voronoi) mesh vertices
        edge_mesh: The edge mesh.
        voronoi_polygons: A list of Voronoi polygon vertices. There is one set of
            Voronoi polygon vertices for each mesh site.
    """

    def __init__(
        self,
        sites: Sequence[Tuple[float, float]],
        elements: Sequence[Tuple[int, int, int]],
        boundary_indices: Sequence[int],
        areas: Union[Sequence[float], None] = None,
        dual_sites: Union[Sequence[Tuple[float, float]], None] = None,
        edge_mesh: Union[EdgeMesh, None] = None,
        voronoi_polygons: Union[List[Sequence[Tuple[float, float]]], None] = None,
    ):
        self.sites = np.asarray(sites).squeeze()
        # Setting dtype to int64 is important when running on Windows.
        # Using default dtype uint64 does not work as Scipy indices in some
        # instances.
        self.elements = np.asarray(elements, dtype=np.int64)
        self.boundary_indices = np.asarray(boundary_indices, dtype=np.int64)
        if areas is not None:
            areas = np.asarray(areas)
        if dual_sites is not None:
            dual_sites = np.asarray(dual_sites)
        self.areas = areas
        self.dual_sites = dual_sites
        self.edge_mesh = edge_mesh
        self.voronoi_polygons = voronoi_polygons

    @property
    def x(self) -> np.ndarray:
        """The x-coordinates of the mesh sites."""
        return self.sites[:, 0]

    @property
    def y(self) -> np.ndarray:
        """The y-coordinates of the mesh sites."""
        return self.sites[:, 1]

    def closest_site(self, xy: Tuple[float, float]) -> int:
        """Returns the index of the mesh site closest to ``(x, y)``.

        Args:
            xy: A shape ``(2, )`` or ``(2, 1)`` sequence of floats, ``(x, y)``.

        Returns:
            The index of the mesh site closest to ``(x, y)``.
        """
        return np.argmin(np.linalg.norm(self.sites - np.atleast_2d(xy), axis=1))

    @staticmethod
    def from_triangulation(
        sites: Sequence[Tuple[float, float]],
        elements: Sequence[Tuple[int, int, int]],
        create_submesh: bool = True,
    ) -> "Mesh":
        """Create a triangular mesh from the coordinates of the triangle vertices
        and a list of indices corresponding to the vertices that connect to triangles.

        Args:
            sites: The (x, y) coordinates of the mesh sites.
            elements: A list of triplets that correspond to the indices of the vertices
                that form a triangle.   E.g. [[0, 1, 2], [0, 1, 3]] corresponds to a
                triangle connecting vertices 0, 1, and 2 and another triangle
                connecting vertices 0, 1, and 3.
            create_submesh: Whether to generate the corresponding
                :class:`tdgl.finit_volume.EdgeMesh` and Voronoi dual mesh.

        Returns:
            A new :class:`tdgl.finite_volume.Mesh` instance
        """
        sites = np.asarray(sites).squeeze()
        elements = np.asarray(elements).squeeze()
        if sites.ndim != 2 or sites.shape[1] != 2:
            raise ValueError(
                f"The site coordinates must have shape (n, 2), got {sites.shape!r}"
            )
        if elements.ndim != 2 or elements.shape[1] != 3:
            raise ValueError(
                f"The elements must have shape (m, 3), got {elements.shape!r}."
            )
        boundary_indices = Mesh.find_boundary_indices(elements)
        dual_sites = edge_mesh = polygons = areas = None
        if create_submesh:
            dual_sites = generate_voronoi_vertices(sites, elements)
            edge_mesh = EdgeMesh.from_mesh(sites, elements, dual_sites)
            areas, polygons = Mesh.compute_voronoi_areas_polygons(
                sites, elements, dual_sites, edge_mesh, boundary_indices
            )
        return Mesh(
            sites=sites,
            elements=elements,
            boundary_indices=boundary_indices,
            edge_mesh=edge_mesh,
            voronoi_polygons=polygons,
            dual_sites=dual_sites,
            areas=areas,
        )

    @staticmethod
    def find_boundary_indices(elements: np.ndarray) -> np.ndarray:
        """Find the boundary vertices.

        Args:
            elements: The triangular elements.

        Returns:
            An array of site indices corresponding to the boundary.
        """
        edges, is_boundary = get_edges(elements)
        # Get the boundary edges and all boundary points
        boundary_edges = edges[is_boundary]
        return np.unique(boundary_edges.flatten())

    @staticmethod
    def compute_voronoi_areas_polygons(
        sites: np.ndarray,
        elements: np.ndarray,
        dual_sites: np.ndarray,
        edge_mesh: EdgeMesh,
        boundary_indices: np.ndarray,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Compute the area and indices of the Voronoi region for each vertex.

        Args:
            sites: The (x, y) coordinates of the mesh sites.
            elements: The mesh triangle indices.
            dual_sites: The (x, y) coordinates of the dual mesh vertices.
            edge_mesh: A :class:`tdgl.finite_volume.EdgeMesh` instance for the
                triangulation defined by ``sites`` and ``elements``.
            boundary_indices: The site indices corresponding to the boundary.

        Returns:
            The Voronoi cell areas and the counterclockwise-oriented vertices
            of the Voronoi cells.
        """
        # Compute polygons to use when computing area
        polygon_indices = get_voronoi_polygon_indices(elements, len(sites))
        # Get the areas for each vertex
        areas, voronoi_polygons = compute_voronoi_polygon_areas(
            sites=sites,
            dual_sites=dual_sites,
            boundary=boundary_indices,
            edges=edge_mesh.edges,
            boundary_edge_indices=edge_mesh.boundary_edge_indices,
            polygons=polygon_indices,
        )
        return areas, voronoi_polygons

    def get_quantity_on_site(
        self, quantity_on_edge: np.ndarray, vector: bool = True
    ) -> np.ndarray:
        """Compute the quantity on site by averaging over all edges
        connecting to each site.

        Args:
            quantity_on_edge: Observable on the edges.
            vector: Whether ``quantity_on_edge`` is a vector quantity.

        Returns:
            The quantity vector or scalar at each site.
        """
        # Normalize the edge direction
        directions = self.edge_mesh.directions
        normalized_directions = (
            directions / np.linalg.norm(directions, axis=1)[:, np.newaxis]
        )
        if vector:
            flux_x = quantity_on_edge * normalized_directions[:, 0]
            flux_y = quantity_on_edge * normalized_directions[:, 1]
        else:
            flux_x = flux_y = quantity_on_edge
        # Sum x and y components for every edge connecting to the vertex
        vertices = np.concatenate(
            [self.edge_mesh.edges[:, 0], self.edge_mesh.edges[:, 1]]
        )
        x_values = np.concatenate([flux_x, flux_x])
        y_values = np.concatenate([flux_y, flux_y])
        counts = np.bincount(vertices)
        x_group_values = np.bincount(vertices, weights=x_values) / counts
        y_group_values = np.bincount(vertices, weights=y_values) / counts
        vector_val = np.array([x_group_values, y_group_values]).T / 2
        if vector:
            return vector_val
        return vector_val[:, 0]

    def smooth(self, iterations: int, create_submesh: bool = True) -> "Mesh":
        """Perform Laplacian smoothing of the mesh, i.e., moving each interior vertex
        to the arithmetic average of its neighboring points.

        Args:
            iterations: The number of smoothing iterations to perform.
            create_submesh: Whether to create the dual mesh and edge mesh.

        Returns:
            A new :class:`tdgl.finite_volume.Mesh` with relaxed vertex positions.
        """
        mesh = self
        elements = mesh.elements
        edges, _ = get_edges(elements)
        n = len(mesh.sites)
        shape = (n, 2)
        boundary = mesh.boundary_indices
        for i in range(iterations):
            sites = mesh.sites
            num_neighbors = np.bincount(edges.ravel(), minlength=shape[0])

            new_sites = np.zeros(shape)
            vals = sites[edges[:, 1]].T
            new_sites += np.array(
                [np.bincount(edges[:, 0], val, minlength=n) for val in vals]
            ).T
            vals = sites[edges[:, 0]].T
            new_sites += np.array(
                [np.bincount(edges[:, 1], val, minlength=n) for val in vals]
            ).T
            new_sites /= num_neighbors[:, np.newaxis]
            # reset boundary points
            new_sites[boundary] = sites[boundary]
            mesh = Mesh.from_triangulation(
                new_sites,
                elements,
                create_submesh=(create_submesh and (i == (iterations - 1))),
            )
        return mesh

    def plot(
        self,
        ax: Union[plt.Axes, None] = None,
        show_sites: bool = True,
        show_edges: bool = False,
        show_dual_edges: bool = True,
        show_voronoi_centroids: bool = False,
        site_color: Union[str, Sequence[float], None] = None,
        edge_color: Union[str, Sequence[float], None] = "k",
        centroid_color: Union[str, Sequence[float], None] = None,
        dual_edge_color: Union[str, Sequence[float], None] = "k",
        linewidth: float = 0.75,
        linestyle: str = "-",
        marker: str = ".",
    ) -> plt.Axes:
        """Plot the mesh.

        Args:
            ax: A :class:`plt.Axes` instance on which to plot the mesh.
            show_sites: Whether to show the mesh sites.
            show_edges: Whether to show the mesh edges.
            show_dual_edges: Whether to show the dual mesh edges.
            show_voronoi_centroids: Whether to show the centroid of each Voronoi cell.
            site_color: The color for the sites.
            edge_color: The color for the edges.
            dual_edge_color: The color for the dual edges.
            centroid_color: The color for the Voronoi centroids.
            linewidth: The line width for all edges.
            linestyle: The line style for all edges.
            marker: The marker to use for the mesh sites and Voronoi centroids.

        Returns:
            The resulting :class:`plt.Axes`
        """

        if ax is None:
            _, ax = plt.subplots()
        ax.set_aspect("equal")

        x, y = self.sites.T
        tri = self.elements

        if show_edges:
            ax.triplot(x, y, tri, color=edge_color, ls=linestyle, lw=linewidth)
        if show_dual_edges:
            for poly in self.voronoi_polygons:
                ax.plot(
                    *close_curve(poly).T,
                    color=dual_edge_color,
                    ls=linestyle,
                    lw=linewidth,
                )
        if show_sites:
            ax.plot(x, y, marker=marker, ls="", color=site_color)
        if show_voronoi_centroids:
            centroids = [convex_polygon_centroid(p) for p in self.voronoi_polygons]
            ax.plot(*np.array(centroids).T, marker=marker, ls="", color=centroid_color)

        return ax

    def to_hdf5(self, h5group: h5py.Group, compress: bool = False) -> None:
        """Save the mesh to a :class:`h5py.Group`.

        Args:
            h5group: The :class:`h5py.Group` into which to store the mesh.
            compress: If ``True``, store only the sites and elements.
        """
        h5group["sites"] = self.sites
        h5group["elements"] = self.elements
        if not compress:
            h5group["boundary_indices"] = self.boundary_indices
            h5group["areas"] = self.areas
            self.edge_mesh.to_hdf5(h5group.create_group("edge_mesh"))
            if self.dual_sites is not None:
                h5group["dual_sites"] = self.dual_sites
            # Save the Voronoi polygon vertices in a single shape (n, 2) array.
            # The ragged list of polygon vertices can be recovered by calling
            # np.split(polygons_flat, split_indices)
            split_indices = np.cumsum(
                [len(polygon) for polygon in self.voronoi_polygons[:-1]]
            )
            polygons_flat = np.concatenate(self.voronoi_polygons, axis=0)
            h5group["voronoi_polygons_flat"] = polygons_flat
            h5group["voronoi_split_indices"] = split_indices

    @staticmethod
    def from_hdf5(h5group: h5py.Group) -> "Mesh":
        """Load a mesh from an HDF5 file.

        Args:
            h5group: The HDF5 group to load the mesh from.

        Returns:
            The loaded mesh.
        """
        if not ("sites" in h5group and "elements" in h5group):
            raise IOError("Could not load mesh due to missing data.")

        if Mesh.is_restorable(h5group):
            polygons_flat = np.array(h5group["voronoi_polygons_flat"])
            voronoi_indices = np.array(h5group["voronoi_split_indices"])
            voronoi_polygons = np.split(polygons_flat, voronoi_indices)
            return Mesh(
                sites=np.array(h5group["sites"]),
                elements=np.array(h5group["elements"], dtype=np.int64),
                boundary_indices=np.array(h5group["boundary_indices"], dtype=np.int64),
                areas=np.array(h5group["areas"]),
                dual_sites=np.array(h5group["dual_sites"]),
                voronoi_polygons=voronoi_polygons,
                edge_mesh=EdgeMesh.from_hdf5(h5group["edge_mesh"]),
            )
        # Recreate mesh from triangulation data if not all data is available
        return Mesh.from_triangulation(
            sites=np.array(h5group["sites"]).squeeze(),
            elements=np.array(h5group["elements"]),
        )

    @staticmethod
    def is_restorable(h5group: h5py.Group) -> bool:
        """Returns ``True`` if the :class:`h5py.Group` contains all of the data
        necessary to create a :class:`tdgl.finite_volume.Mesh` without re-computing
        any values.

        Args:
            h5group: The :class:`h5py.Group` to check.

        Returns:
            Whether the mesh can be restored from the given group.
        """
        return (
            "sites" in h5group
            and "elements" in h5group
            and "boundary_indices" in h5group
            and "areas" in h5group
            and "edge_mesh" in h5group
            and "dual_sites" in h5group
            and "voronoi_polygons_flat" in h5group
            and "voronoi_split_indices" in h5group
        )



from typing import Dict, List, NamedTuple, Optional, Union

import numpy as np
import pint

from .device.device import Device


class Fluxoid(NamedTuple):
    """The fluxoid for a closed region :math:`S` with boundary :math:`\\partial S`
    is defined as:

    .. math::

        \\begin{split}
        \\Phi^f_S &= \\Phi^f_{S,\\text{ flux}} + \\Phi^f_{S,\\text{ supercurrent}}
        \\\\&=\\int_S \\mu_0 H_z(\\mathbf{r})\\,\\mathrm{d}^2r +
            \\oint_{\\partial S}
            \\mu_0\\Lambda(\\mathbf{r})\\mathbf{K}_s(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r}
        \\\\&=\\oint_{\\partial S} \\mathbf{A}(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r} +
            \\oint_{\\partial S}
            \\mu_0\\Lambda(\\mathbf{r})\\mathbf{K}_s(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r}
        \\end{split}

    Args:
        flux_part: :math:`\\int_S \\mu_0 H_z(\\mathbf{r})\\,\\mathrm{d}^2r=\\oint_{\\partial S}\\mathbf{A}(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r}`.
        supercurrent_part: :math:`\\oint_{\\partial S}\\mu_0\\Lambda(\\mathbf{r})\\mathbf{K}_s(\\mathbf{r})\\cdot\\mathrm{d}\\mathbf{r}`.
    """

    flux_part: Union[float, pint.Quantity]
    supercurrent_part: Union[float, pint.Quantity]


def make_fluxoid_polygons(
    device: Device,
    holes: Optional[Union[List[str], str]] = None,
    join_style: str = "mitre",
    interp_points: Optional[int] = None,
) -> Dict[str, np.ndarray]:
    """Generates polygons enclosing the given holes to calculate the fluxoid.

    Args:
        device: The Device for which to generate polygons.
        holes: Name(s) of the hole(s) in the device for which to generate polygons.
            Defaults to all holes in the device.
        join_style: See :meth:`tdgl.Polygon.buffer`.
        interp_points: If provided, the resulting polygons will be interpolated to
            ``interp_points`` vertices.

    Returns:
        A dict of ``{hole_name: fluxoid_polygon}``.
    """
    device_polygons = [device.film] + device.holes
    device_holes = {hole.name: hole for hole in device.holes}
    if holes is None:
        holes = list(device_holes)
    if isinstance(holes, str):
        holes = [holes]
    polygons = {}
    for name in holes:
        hole = device_holes[name]
        hole_poly = hole.polygon
        min_dist = min(
            hole_poly.exterior.distance(other.polygon.exterior)
            for other in device_polygons
            if other.name != name
        )
        delta = min_dist / 2
        new_poly = hole.buffer(delta, join_style=join_style)
        if interp_points:
            new_poly = new_poly.resample(interp_points)
        polygons[name] = new_poly.points
    return polygons



import pytest

import tdgl
from tdgl.geometry import box, circle


@pytest.fixture(scope="package")
def transport_device():
    london_lambda = 2
    xi = 1
    d = 0.1

    layer = tdgl.Layer(coherence_length=xi, london_lambda=london_lambda, thickness=d)
    film = (
        tdgl.Polygon("film", points=box(10)).union(box(30, 4, points=400)).resample(501)
    )
    hole = tdgl.Polygon("hole1", points=circle(1.5, center=(2, 2)))
    source = tdgl.Polygon(points=box(1e-2, 4, center=(-15, 0))).set_name("source")
    drain = source.scale(xfact=-1).set_name("drain")

    device = tdgl.Device(
        "film",
        layer=layer,
        film=film,
        holes=[hole, hole.scale(xfact=-1, yfact=-1).set_name("hole2")],
        terminals=[source, drain],
        probe_points=[(-10, 0), (10, 0)],
    )

    assert device.mesh is None
    assert device.points is None
    assert device.triangles is None
    assert device.edges is None
    assert device.edge_lengths is None
    assert device.areas is None
    assert device.probe_point_indices is None
    assert device.boundary_sites() is None

    _ = device.mesh_stats_dict()
    _ = device.mesh_stats()

    device.make_mesh(min_points=2000, smooth=100, max_edge_length=xi / 2)

    _ = device.areas
    _ = device.boundary_sites()
    _ = device.mesh_stats_dict()
    _ = device.mesh_stats()

    return device


@pytest.fixture(scope="package")
def transport_device_solution(transport_device):
    device = transport_device
    dt = 1e-3
    total_time = 100

    options = tdgl.SolverOptions(
        dt_init=dt,
        solve_time=total_time,
        save_every=100,
        field_units="uT",
        current_units="uA",
        include_screening=False,
    )
    field = tdgl.sources.ConstantField(10)
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=field,
        terminal_currents=dict(source=10, drain=-10),
    )
    return solution


@pytest.fixture(scope="package")
def box_device():
    london_lambda = 1.5
    xi = 1.5
    d = 0.1
    layer = tdgl.Layer(coherence_length=xi, london_lambda=london_lambda, thickness=d)
    film = tdgl.Polygon("film", points=box(10))
    device = tdgl.Device("film", layer=layer, film=film)
    device.make_mesh(min_points=2000, smooth=40, max_edge_length=xi / 2)
    return device


@pytest.fixture(scope="package")
def box_device_solution_no_screening(box_device):
    device = box_device
    dt = 1e-3
    total_time = 20

    options = tdgl.SolverOptions(
        dt_init=dt,
        solve_time=total_time,
        save_every=100,
        field_units="uT",
        current_units="uA",
        include_screening=False,
    )
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=tdgl.sources.ConstantField(50),
        terminal_currents=None,
    )
    return solution


@pytest.fixture(scope="package")
def box_device_solution_with_screening(box_device):
    device = box_device
    dt = 1e-3
    total_time = 20

    options = tdgl.SolverOptions(
        dt_init=dt,
        solve_time=total_time,
        save_every=100,
        field_units="uT",
        current_units="uA",
        include_screening=True,
    )
    solution = tdgl.solve(
        device,
        options,
        applied_vector_potential=tdgl.sources.ConstantField(50),
        terminal_currents=None,
    )
    return solution



import inspect
import os
import sys
import time
from typing import Dict, Optional

import IPython
import joblib
import matplotlib
import numpy
import scipy
from IPython.display import HTML

try:
    import jax

    jax_version = jax.__version__
except (ModuleNotFoundError, ImportError):
    jax_version = None

import tdgl


def _blas_info() -> str:
    # https://github.com/qutip/qutip/blob/3ff3f42d2cd950c99af8936b12d9221ed79de5e3/
    # qutip/utilities.py#L335-L355
    config = numpy.__config__
    if hasattr(config, "blas_ilp64_opt_info"):
        blas_info = config.blas_ilp64_opt_info
    elif hasattr(config, "blas_opt_info"):
        blas_info = config.blas_opt_info
    else:
        blas_info = {}
    _has_lib_key = "libraries" in blas_info
    blas = None
    if hasattr(config, "mkl_info") or (
        _has_lib_key and any("mkl" in lib for lib in blas_info["libraries"])
    ):
        blas = "INTEL MKL"
    elif hasattr(config, "openblas_info") or (
        _has_lib_key and any("openblas" in lib for lib in blas_info["libraries"])
    ):
        blas = "OPENBLAS"
    elif "extra_link_args" in blas_info and (
        "-Wl,Accelerate" in blas_info["extra_link_args"]
    ):
        blas = "Accelerate"
    else:
        blas = "Generic"
    return blas


def version_dict() -> Dict[str, str]:
    """Returns a dictionary containing the versions of important dependencies."""
    cpu_count = [joblib.cpu_count(only_physical_cores=b) for b in (True, False)]
    return {
        "tdgl": tdgl.__version__,
        "Numpy": numpy.__version__,
        "SciPy": scipy.__version__,
        "matplotlib": matplotlib.__version__,
        "jax": str(jax_version),
        "IPython": IPython.__version__,
        "Python": sys.version,
        "OS": f"{os.name} [{sys.platform}]",
        "Number of CPUs": f"Physical: {cpu_count[0]}, Logical: {cpu_count[1]}",
        "BLAS Info": _blas_info(),
    }


def version_table(
    version_info: Optional[Dict[str, str]] = None, verbose: bool = False
) -> HTML:
    """Returns an HTML table with the versions of important depedencies."""

    # Adapted from: https://github.com/qutip/qutip/blob/
    # 88919ce50880dadbc1a817a3e6059c82c23a83f9/qutip/ipynbtools.py#L47

    html = [
        "<table>",
        "<tr><th>Software</th><th>Version</th></tr>",
    ]
    if version_info is None:
        version_info = version_dict()

    for name, version in version_info.items():
        html.append(f"<tr><td>{name}</td><td>{version}</td></tr>")

    if verbose:
        html.append("<tr><th colspan='2'>Additional information</th></tr>")
        install_path = os.path.dirname(inspect.getsourcefile(tdgl))
        html.append(f"<tr><td>Installation path</td><td>{install_path}</td></tr>")

    html.append(
        f"<tr><td colspan='2'>{time.strftime('%a %b %d %H:%M:%S %Y %Z')}</td></tr>"
    )
    html.append("</table>")

    return HTML("".join(html))
